import logging
from string import Template
from typing import Union

from magnus import defaults, utils

logger = logging.getLogger(defaults.NAME)
try:
    import docker
except ImportError:
    logger.info('docker was not installed, docker functionality will not work')


def generate_docker_file(style: str = "poetry", git_tracked: bool = True):
    install_style, install_requirements, copy_content = None, None, None

    if style == 'poetry':
        install_style = 'RUN pip install poetry'
        install_requirements = 'RUN poetry install'
        logger.info('Using poetry style for requirements')
    elif style == 'pipenv':
        install_style = 'RUN pip install pipenv'
        install_requirements = 'RUN pipenv install'
        logger.info('Using pipenv style for requirements')
    else:
        install_requirements = 'RUN pip install -r requirements.txt'
        logger.info('Trying requirements.txt, if one exists')

    copy_content = 'COPY . /app'
    if git_tracked:
        if not utils.is_a_git_repo():
            msg = (
                "The current project is not git versioned. Disable only git tracked to create the image. "
                "Be aware over-riding this can cause leak of sensitive data if you are not careful."
            )
            raise Exception(msg)
        copy_content = 'ADD git_tracked.tar.gz /app'
        utils.archive_git_tracked(defaults.GIT_ARCHIVE_NAME)

    dockerfile_content = Template(defaults.DOCKERFILE_CONTENT).safe_substitute(
        {'INSTALL_STYLE': install_style, 'INSTALL_REQUIREMENTS': install_requirements,
         'COPY_CONTENT': copy_content})
    with open(defaults.DOCKERFILE_NAME, 'w', encoding='utf-8') as fw:
        fw.write(dockerfile_content)


def build_docker(image_name: str, docker_file: Union[str, None],
                 style: str, tag: str, commit_tag: bool, dry_run: bool = False, git_tracked: bool = True):
    if commit_tag:
        if not utils.is_a_git_repo():
            msg = (
                "The current project is not git versioned, cannot use commit tag option when building image"
            )
            raise Exception(msg)

        tag = utils.get_current_code_commit()[:defaults.LEN_SHA_FOR_TAG]  # type: ignore

    if not docker_file:
        generate_docker_file(style, git_tracked=git_tracked)
        docker_file = defaults.DOCKERFILE_NAME

    if dry_run:
        return

    docker_client = docker.from_env()
    docker_client.images.build(path=f".", dockerfile=docker_file, tag=f'{image_name}:{tag}', quiet=False)



from __future__ import annotations

import hashlib
import json
import logging
import os
import subprocess
import sys
from collections import OrderedDict
from datetime import datetime
from inspect import signature
from pathlib import Path
from string import Template as str_template
from types import FunctionType
from typing import TYPE_CHECKING, Callable, List, Mapping, Tuple, Union

from ruamel.yaml import YAML  # type: ignore
from stevedore import driver

from magnus import defaults

if TYPE_CHECKING:
    from magnus.executor import BaseExecutor
    from magnus.nodes import BaseNode


logger = logging.getLogger(defaults.NAME)

try:
    import docker
except ImportError:
    logger.warning('docker was not installed, local-container mode will not function')


def does_file_exist(file_path: str) -> bool:
    """
    Check if a file exists.
    Implemented here to avoid repetition of logic.

    Args:
        file_path (str): The file path to check

    Returns:
        bool: False if it does not otherwise True
    """
    my_file = Path(file_path)
    return my_file.is_file()


def does_dir_exist(file_path: Union[str, Path]) -> bool:
    """
    Check if a directory exists.
    Implemented here to avoid repetition of logic

    Args:
        file_path (str or Path): The directory path to check

    Returns:
        bool: False if the directory does not exist, True otherwise
    """
    my_file = Path(file_path)
    return my_file.is_dir()


def safe_make_dir(directory: Union[str, Path]):
    """
    Safely make the directory.
    Ignore if it exists and create the parents if necessary.

    Args:
        directory (str): The directory path to create
    """
    Path(directory).mkdir(parents=True, exist_ok=True)


def generate_run_id(run_id: str = None) -> str:
    """
    Generate a new run_id.

    If the input run_id is none, we create one based on time stamp.

    Args:
        run_id (str, optional): Input Run ID. Defaults to None

    Returns:
        str: A generated run_id
    """
    # If we are not provided with a run_id, generate one
    if not run_id:
        run_id = datetime.now().strftime('%Y%m%d%H%M%S')

    return run_id


def apply_variables(apply_to: dict, variables: dict) -> dict:
    """
    Safely applies the variables to a config.

    For example: For config:
         {'a' : ${b}}, the value of ${b} is replaced by b in the variables.

    If the ${b} does not exist in the variables, it is ignored in the config.

    Args:
        apply_to (dict): The config to apply variables
        variables (dict): The variables in key, value pairs

    Raises:
        Exception: If the variables is not dict

    Returns:
        dict: A transformed dict with variables applied
    """
    if not isinstance(variables, dict):
        raise Exception('Argument Variables should be dict')

    json_d = json.dumps(apply_to)
    transformed = str_template(json_d).safe_substitute(**variables)
    return json.loads(transformed)


def get_module_and_func_names(command: str) -> Tuple[str, str]:
    """
    Given a string of module.function, this functions returns the module name and func names.

    It also checks to make sure that the string is of expected 'module.func' format

    Args:
        command (str): String of format module.function_name

    Raises:
        Exception: If the string is of not format

    Returns:
        Tuple[str, str]: (module_name, function_name) extracted from the input string
    """
    mods = command.split('.')
    if len(mods) <= 1:
        raise Exception('The command should be a function to call')
    func = mods[-1]
    module = '.'.join(mods[:-1])
    return module, func


def get_module_and_func_from_function(command: FunctionType) -> str:
    module_name = sys.modules[command.__module__]
    func_name = command.__name__
    return f"{module_name}.{func_name}"


def get_dag_hash(dag: dict) -> str:
    """
    Generates the hash of the dag definition

    Args:
        dag (dict): The dictionary object containing the dag definition

    Returns:
        str: The hash of the dag definition
    """
    dag_str = json.dumps(dag, sort_keys=True, ensure_ascii=True)
    return hashlib.sha1(dag_str.encode('utf-8')).hexdigest()


def load_yaml(file_path: str, load_type: str = 'safe') -> dict:
    """
    Loads an yaml and returns the dictionary

    Args:
        file_path (str): The path of the yamlfile
        load_type (str, optional): The load type as understood by ruamel. Defaults to 'safe'.

    Returns:
        dict: The mapping as defined in the yaml file
    """
    with open(file_path, encoding='utf-8') as f:
        yaml = YAML(typ=load_type, pure=True)
        yaml_config = yaml.load(f)
    return yaml_config


def is_a_git_repo() -> bool:
    """
    Does a git command to see if the project is git versioned

    Returns:
        bool: True if it is git versioned, False otherwise
    """
    command = 'git rev-parse --is-inside-work-tree'
    try:
        subprocess.check_output(
            command.split()).strip().decode('utf-8')
        logger.info('Found the code to be git versioned')
        return True
    except BaseException:  # pylint: disable=W0702
        logger.error('No git repo found, unsafe hash')

    return False


def get_current_code_commit() -> Union[str, None]:
    """
    Gets the git sha id if the project is version controlled

    Returns:
        Union[str, None]: SHA ID if the code is versioned, None otherwise
    """
    command = 'git rev-parse HEAD'
    if not is_a_git_repo():
        return None
    try:
        label = subprocess.check_output(
            command.split()).strip().decode('utf-8')
        logger.info("Found the git commit to be: %s", label)
        return label
    except BaseException:  # pylint: disable=W0702
        logger.exception('Error getting git hash')
        raise


def archive_git_tracked(name: str):
    command = f'git archive -v -o {name}.tar.gz --format=tar.gz HEAD'

    if not is_a_git_repo():
        raise Exception("Not a git repo")
    try:
        subprocess.check_output(
            command.split()).strip().decode('utf-8')
    except BaseException:  # pylint: disable=W0702
        logger.exception('Error archiving repo')
        raise


def is_git_clean() -> Tuple[bool, Union[None, str]]:
    """
    Checks if the git tree is clean and there are no modified tracked files

    Returns:
        Tuple[bool, Union[None, str]]: None if its clean, comma-seperated file names if it is changed
    """
    command = 'git diff --name-only'
    if not is_a_git_repo():
        return False, None
    try:
        label = subprocess.check_output(
            command.split()).strip().decode('utf-8')
        if not label:
            return True, None
        return False, label
    except BaseException:  # pylint: disable=W0702
        logger.exception('Error checking if the code is git clean')

    return False, None


def get_git_remote() -> Union[str, None]:
    """
    Gets the remote URL of git

    Returns:
        Union[str, None]: Remote URL if the code is version controlled, None otherwise
    """
    command = 'git config --get remote.origin.url'
    if not is_a_git_repo():
        return None
    try:
        label = subprocess.check_output(
            command.split()).strip().decode('utf-8')
        logger.info('Found the git remote to be: %s', label)
        return label
    except BaseException:  # pylint: disable=W0702
        logger.exception('Error getting git remote')
        raise


def get_local_docker_image_id(image_name: str) -> str:
    """
    If we are running in local settings, return the docker image id

    Args:
        image_name (str): The image name we need the digest for

    Returns:
        str: The docker image digest
    """
    try:
        client = docker.from_env()
        image = client.images.get(image_name)
        return image.attrs['Id']
    except ImportError:  # pragma: no cover
        logger.warning('Did not find docker installed, some functionality might be affected')
    except BaseException:
        logger.exception(f'Could not find the image by name {image_name}')

    return ''


def get_git_code_identity(run_log_store):
    """
    Returns a code identity object for version controlled code.

    Args:
        run_log_store (magnus.datastore.BaseRunLogStore): The run log store used in this process

    Returns:
        magnus.datastore.CodeIdentity: The code identity used by the run log store.
    """
    code_identity = run_log_store.create_code_identity()
    try:
        code_identity.code_identifier = get_current_code_commit()
        code_identity.code_identifier_type = 'git'
        code_identity.code_identifier_dependable, changed = is_git_clean()
        code_identity.code_identifier_url = get_git_remote()
        if changed:
            code_identity.code_identifier_message = 'changes found in ' + ', '.join(changed.split('\n'))
    except BaseException:
        logger.exception("Git code versioning problems")

    return code_identity


def remove_prefix(text: str, prefix: str) -> str:
    """
    Removes a prefix if one is present in the input text

    Args:
        text (str): The input text to remove the prefix from
        prefix (str): The prefix that has to be removed

    Returns:
        str: The original string if no prefix is found, or the right prefix chomped string if present
    """
    if text.startswith(prefix):
        return text[len(prefix):]
    return text  # or whatever is given


def get_tracked_data() -> dict:
    """
    Scans the environment variables to find any user tracked variables that have a prefix MAGNUS_TRACK_
    Removes the environment variable to prevent any clashes in the future steps

    Returns:
        dict: A dictionary of user tracked data
    """
    tracked_data = {}
    for env_var, value in os.environ.items():
        if env_var.startswith(defaults.TRACK_PREFIX):
            key = remove_prefix(env_var, defaults.TRACK_PREFIX)
            tracked_data[key.lower()] = json.loads(value)
            del os.environ[env_var]
    return tracked_data


def get_user_set_parameters(remove: bool = False) -> dict:
    """
    Scans the environment variables for any user returned parameters that have a prefix MAGNUS_PRM_

    Args:
        remove (bool, optional): Flag to remove the parameter if needed. Defaults to False.

    Returns:
        dict: The dictionary of found user returned parameters
    """
    parameters = {}
    for env_var, value in os.environ.items():
        if env_var.startswith(defaults.PARAMETER_PREFIX):
            key = remove_prefix(env_var, defaults.PARAMETER_PREFIX)
            try:
                parameters[key.lower()] = json.loads(value)
            except json.decoder.JSONDecodeError:
                logger.error(f'Parameter {key} could not be JSON decoded, adding the literal value')
                parameters[key.lower()] = value

            if remove:
                del os.environ[env_var]
    return parameters


def hash_bytestr_iter(bytesiter, hasher, ashexstr=True):  # pylint: disable=C0116
    for block in bytesiter:  # pragma: no cover
        hasher.update(block)
    return hasher.hexdigest() if ashexstr else hasher.digest()  # pragma: no cover


def file_as_blockiter(afile, blocksize=65536):  # pylint: disable=C0116
    with afile:  # pragma: no cover
        block = afile.read(blocksize)
        while len(block) > 0:
            yield block
            block = afile.read(blocksize)


def get_data_hash(file_name: str):
    """
    Returns the hash of the data file

    Args:
        file_name (str): The file name to generated the hash

    Returns:
        str: The SHA ID of the file contents
    """
    # https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
    return hash_bytestr_iter(file_as_blockiter(open(file_name, 'rb')), hashlib.sha256())  # pragma: no cover


def filter_arguments_for_func(func: Callable, parameters: dict, map_variable: dict = None) -> dict:
    """
    Inspects the function to be called as part of the pipeline to find the arguments of the function.
    Matches the function arguments to the parameters available either by command line or by up stream steps.

    Args:
        func (Callable): The function to inspect
        parameters (dict): The parameters available for the run

    Returns:
        dict: The parameters matching the function signature
    """
    sign = signature(func)
    return filter_arguments_from_parameters(parameters=parameters,
                                            signature_parameters=sign.parameters,
                                            map_variable=map_variable)


def filter_arguments_from_parameters(
        parameters: dict, signature_parameters: Union[List, Mapping],
        map_variable: dict = None) -> dict:
    arguments = {}

    for param, value in parameters.items():
        if param in signature_parameters:
            arguments[param] = value

    if map_variable:
        for iterate_as, value in map_variable.items():
            if iterate_as in signature_parameters:
                arguments[iterate_as] = value

    return arguments


def get_node_execution_command(
        executor: BaseExecutor, node: BaseNode, map_variable: dict = None, over_write_run_id: str = '') -> str:
    """
    A utility function to standardize execution call to a node via command line.

    Args:
        executor (object): The executor class.
        node (object): The Node to execute
        map_variable (str, optional): If the node belongs to a map step. Defaults to None.

    Returns:
        str: The execution command to run a node via command line.
    """
    run_id = executor.run_id

    if over_write_run_id:
        run_id = over_write_run_id

    log_level = logging.getLevelName(logger.getEffectiveLevel())

    action = (f'magnus execute_single_node {run_id} '
              f'{node._command_friendly_name()}'
              f' --log-level {log_level}'
              )
    if executor.pipeline_file:
        action = action + f" --file {executor.pipeline_file}"

    if map_variable:
        action = action + f" --map-variable '{json.dumps(map_variable)}'"

    if executor.configuration_file:
        action = action + f' --config-file {executor.configuration_file}'

    if executor.parameters_file:
        action = action + f' --parameters-file {executor.parameters_file}'

    if executor.tag:
        action = action + f' --tag {executor.tag}'

    return action


def get_service_namespace(service_type: str) -> str:
    """
    Return the namespace of the service type

    Args:
        service_type (str): Should be one of executor, run_log_store, catalog, secrets

    Raises:
        Exception: If the service type is not one of the accepted values

    Object:
        str: The name space of the base class
    """
    if service_type == 'executor':
        return 'magnus.executor.BaseExecutor'

    if service_type == 'run_log_store':
        return 'magnus.datastore.BaseRunLogStore'

    if service_type == 'catalog':
        return 'magnus.catalog.BaseCatalog'

    if service_type == 'secrets':
        return 'magnus.secrets.BaseSecrets'

    if service_type == 'experiment_tracking':
        return 'magnus.experiment_tracker.BaseExperimentTracker'

    raise Exception('Service type is not recognized')


def get_provider_by_name_and_type(service_type: str, service_details: dict):
    """
    Given a service type, one of executor, run_log_store, catalog, secrets and the config
    return the exact child class implementing the service.
    We use stevedore to do the work for us.

    Args:
        service_type (str): One of executor, run_log_store, catalog, secrets
        service_details (dict): The config used to instantiate the service.

    Raises:
        Exception: If the service by that name does not exist

    Returns:
        object: A service object
    """
    namespace = get_service_namespace(service_type=service_type)

    service_name = service_details['type']
    service_config = {}
    if 'config' in service_details:
        service_config = service_details.get('config', {})

    logger.info(f'Trying to get a service of {service_type} of the name {service_name} with config: {service_config}')
    try:
        mgr = driver.DriverManager(
            namespace=namespace,
            name=service_name,
            invoke_on_load=True,
            invoke_kwds={'config': service_config}
        )
        return mgr.driver
    except Exception as _e:
        raise Exception(f'Could not find the service of type: {service_type} with config: {service_details}') from _e


def get_duration_between_datetime_strings(start_time: str, end_time: str) -> str:
    """Given two datetime strings, compute the duration between them

    Args:
        start_time (str): ISO format datetime string
        end_time (str): ISO format datetime string
    Returns:
        The duration between the time in string format
    """
    start = datetime.fromisoformat(start_time)
    end = datetime.fromisoformat(end_time)

    return str(end - start)


def get_run_config(executor: BaseExecutor) -> dict:
    """
    Given an executor with assigned services, return the run_config

    Args:
        executor (object): The executor with all the services assigned.

    Returns:
        dict: The run_config.
    """
    run_config = {}

    run_config['executor'] = {'type': executor.service_name,
                              'config': executor.config}

    run_config['run_log_store'] = {'type': executor.run_log_store.service_name,
                                   'config': executor.run_log_store.config}

    run_config['catalog'] = {'type': executor.catalog_handler.service_name,
                             'config': executor.catalog_handler.config}

    run_config['secrets'] = {'type': executor.secrets_handler.service_name,
                             'config': executor.secrets_handler.config}

    run_config['experiment_tracker'] = {'type': executor.experiment_tracker.service_name,
                                        'config': executor.experiment_tracker.config}
    run_config['variables'] = executor.variables  # type: ignore
    run_config['pipeline'] = executor.dag._to_dict()

    return run_config


def json_to_ordered_dict(json_str: str) -> OrderedDict:
    """
    Decode a JSON str into OrderedDict

    Args:
        json_str ([str]): The JSON string to decode

    Returns:
        [OrderedDict]: The decoded OrderedDict
    """

    if json_str and json_str != '{}':
        return json.loads(json_str, object_pairs_hook=OrderedDict)

    return OrderedDict()


def set_magnus_environment_variables(run_id: str = None, configuration_file: str = None, tag: str = None):
    if run_id:
        os.environ[defaults.ENV_RUN_ID] = run_id

    if configuration_file:
        os.environ[defaults.MAGNUS_CONFIG_FILE] = configuration_file

    if tag:
        os.environ[defaults.MAGNUS_RUN_TAG] = tag


def gather_variables():
    variables = {}

    for env_var, value in os.environ.items():
        if env_var.startswith(defaults.VARIABLE_PREFIX):
            key = remove_prefix(env_var, defaults.VARIABLE_PREFIX)
            variables[key] = value

    return variables



from __future__ import annotations

import copy
import json
import logging
import os
import re
from typing import TYPE_CHECKING, Dict, Optional

from pydantic import BaseModel

from magnus import defaults, exceptions, integration, interaction, utils
from magnus.graph import Graph
from magnus.nodes import BaseNode

if TYPE_CHECKING:
    from magnus.catalog import BaseCatalog
    from magnus.datastore import BaseRunLogStore, StepLog
    from magnus.experiment_tracker import BaseExperimentTracker
    from magnus.secrets import BaseSecrets

logger = logging.getLogger(defaults.NAME)


class BaseExecutor:
    """
    The skeleton of an executor class.
    Any implementation of an executor should inherit this class and over-ride accordingly.

    The logic of any dag execution is a play between three methods of this class.

    execute_graph:
        This method is responsible for traversing A graph.
        The core logic is start at the start_at the graph and traverse according to the state of the execution.
        execute_graph hands over the actual execution of the node to self.execute_from_graph

        Helper method: prepare_for_graph_execution would be called prior to calling execute_graph.
            Use it to modify settings if needed.

    execute_from_graph:
        This method is responsible for executing a node.
        But given that a node itself could be a dag in cases of parallel, map and dag, this method handles the cases.
        If the node is of type task, success, fail: we can pretty much execute it and we call self.trigger_job.
        If the node is of type composite: We call the node's execute_as_graph function which internally
        triggers execute_graph in-turn iteratively traverses the graph.


    _execute_node:
        This method is where the actual execution of the work happens.
        This method is already in the compute environment of the mode.
        Use prepare_node_for_execution to adjust settings.
        The base class is given an implementation and in most cases should not be touched.

        Helper method: prepare_for_node_execution would be called prior to calling _execute_node.
            Use it to modify settings if needed

    The above logic holds good when we are in interactive compute mode i.e. local, local-container, local-aws-batch

    But in 3rd party orchestration mode, we might have to render the job specifications and the roles might be different

    Please see the implementations of local, local-container, local-aws-batch to perform interactive compute.
    And demo-renderer to see an example of what a 3rd party executor looks like.
    """
    service_name = ''

    class Config(BaseModel):
        enable_parallel: bool = defaults.ENABLE_PARALLEL
        placeholders: dict = {}

    def __init__(self, config: dict = None):
        # pylint: disable=R0914,R0913
        config = config or {}
        self.config = self.Config(**config)
        # The remaining would be attached later
        # The definition files
        self.pipeline_file = None
        self.variables_file = None
        self.parameters_file = None
        self.configuration_file = None
        # run descriptors
        self.tag: str = ''
        self.run_id: str = ''
        self.single_step: bool = False
        self.variables: Dict[str, str] = {}
        self.use_cached: bool = False
        self.dag: Graph = None  # type: ignore
        self.dag_hash: str = ''
        self.execution_plan: str = ''
        # Services
        self.catalog_handler: BaseCatalog = None  # type: ignore
        self.secrets_handler: BaseSecrets = None  # type: ignore
        self.experiment_tracker: BaseExperimentTracker = None  # type: ignore
        self.run_log_store: BaseRunLogStore = None  # type: ignore
        self.previous_run_log = None

        self.context_step_log: StepLog = None  # type: ignore

    @property
    def step_decorator_run_id(self):
        return os.environ.get('MAGNUS_RUN_ID', None)

    def _is_parallel_execution(self) -> bool:
        """
        Controls the parallelization of branches in map and parallel state.
        Defaults to False and left for the compute modes to decide.

        Returns:
            bool: True if the mode allows parallel execution of branches.
        """
        return self.config.enable_parallel

    def _set_up_run_log(self, exists_ok=False):
        """
        Create a run log and put that in the run log store

        If exists_ok, we allow the run log to be already present in the run log store
        """
        try:
            attempt_run_log = self.run_log_store.get_run_log_by_id(run_id=self.run_id, full=False)
            if attempt_run_log.status in [defaults.FAIL, defaults.SUCCESS]:
                raise Exception(f'The run log by id: {self.run_id} already exists and is {attempt_run_log.status}')

            raise exceptions.RunLogExistsError(self.run_id)
        except exceptions.RunLogNotFoundError:
            pass
        except exceptions.RunLogExistsError:
            if exists_ok:
                return
            raise

        run_log = {}
        run_log['run_id'] = self.run_id
        run_log['tag'] = self.tag
        run_log['use_cached'] = False
        run_log['status'] = defaults.PROCESSING
        run_log['dag_hash'] = self.dag_hash

        parameters = {}
        if self.parameters_file:
            parameters.update(utils.load_yaml(self.parameters_file))

        # Update these with some from the environment variables
        parameters.update(utils.get_user_set_parameters())

        if self.previous_run_log:
            run_log['original_run_id'] = self.previous_run_log.run_id
            # Sync the previous run log catalog to this one.
            self.catalog_handler.sync_between_runs(previous_run_id=self.previous_run_log.run_id, run_id=self.run_id)
            run_log['use_cached'] = True
            parameters.update(self.previous_run_log.parameters)

        run_log = self.run_log_store.create_run_log(**run_log)
        # Any interaction with run log store attributes should happen via API if available.
        self.run_log_store.set_parameters(run_id=self.run_id, parameters=parameters)

        # Update run_config
        run_config = utils.get_run_config(self)
        self.run_log_store.set_run_config(run_id=self.run_id, run_config=run_config)

    def prepare_for_graph_execution(self):
        """
        This method would be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we need not do anything.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_traversal(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_traversal(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_traversal(self, self.secrets_handler)

        self._set_up_run_log()

    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """

        integration.validate(self, self.run_log_store)
        integration.configure_for_execution(self, self.run_log_store)

        integration.validate(self, self.catalog_handler)
        integration.configure_for_execution(self, self.catalog_handler)

        integration.validate(self, self.secrets_handler)
        integration.configure_for_execution(self, self.secrets_handler)

    def _sync_catalog(self, node: BaseNode, step_log: StepLog, stage: str, synced_catalogs=None):
        """
        1). Identify the catalog settings by over-riding node settings with the global settings.
        2). For stage = get:
                Identify the catalog items that are being asked to get from the catalog
                And copy them to the local compute data folder
        3). For stage = put:
                Identify the catalog items that are being asked to put into the catalog
                Copy the items from local compute folder to the catalog
        4). Add the items onto the step log according to the stage

        Args:
            node (Node): The current node being processed
            step_log (StepLog): The step log corresponding to that node
            stage (str): One of get or put

        """
        if stage not in ['get', 'put']:
            msg = (
                'Catalog service only accepts get/put possible actions as part of node execution.'
                f'Sync catalog of the executor: {self.service_name} asks for {stage} which is not accepted'
            )
            raise Exception(msg)

        node_catalog_settings = node._get_catalog_settings()
        if not (node_catalog_settings and stage in node_catalog_settings):
            # Nothing to get/put from the catalog
            return None

        # Local compute data folder over rides the global one
        compute_data_folder = self.catalog_handler.compute_data_folder
        if 'compute_data_folder' in node_catalog_settings and node_catalog_settings['compute_data_folder']:
            compute_data_folder = node_catalog_settings['compute_data_folder']

        data_catalogs = []
        for name_pattern in node_catalog_settings.get(stage) or []:  #  Assumes a list
            data_catalogs = getattr(
                self.catalog_handler, stage)(
                name=name_pattern, run_id=self.run_id, compute_data_folder=compute_data_folder,
                synced_catalogs=synced_catalogs)

        if data_catalogs:
            step_log.add_data_catalogs(data_catalogs)

        return data_catalogs

    def _execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        This is the entry point when we do the actual execution of the function.
        DO NOT Over-ride this function.

        While in interactive mode, we just compute, in 3rd party interactive mode, we call this function from the CLI

        In most cases,
            * We get the corresponding step_log of the node and the parameters.
            * We sync the catalog to GET any data sets that are in the catalog
            * We call the execute method of the node for the actual compute and retry it as many times as asked.
            * If the node succeeds, we get any of the user defined metrics provided by the user.
            * We sync the catalog to PUT any data sets that are in the catalog.

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node is of a map state, map_variable is the value of the iterable.
                        Defaults to None.

        #TODO: The attempts should ideally by handled outside of this function
        """
        max_attempts = node._get_max_attempts()
        attempts = 0
        step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)

        parameters = self.run_log_store.get_parameters(run_id=self.run_id)
        # Set up environment variables for the execution
        # If the key already exists, do not update it to give priority to parameters set by environment variables
        interaction.store_parameter(update=False, **parameters)

        data_catalogs_get = self._sync_catalog(node, step_log, stage='get')

        mock = step_log.mock
        logger.info(f'Trying to execute node: {node.internal_name}, attempt : {attempts}, max_attempts: {max_attempts}')
        while attempts < max_attempts:
            try:
                self.context_step_log = step_log
                attempt_log = node.execute(executor=self, mock=mock,
                                           map_variable=map_variable, **kwargs)
                attempt_log.attempt_number = attempts
                step_log.attempts.append(attempt_log)
                if attempt_log.status == defaults.FAIL:
                    raise Exception()

                step_log.status = defaults.SUCCESS
                step_log.user_defined_metrics = utils.get_tracked_data()
                self.run_log_store.set_parameters(self.run_id, utils.get_user_set_parameters(remove=True))
                break
            except Exception as _e:  # pylint: disable=W0703
                attempts += 1
                logger.exception(f'Node: {node} failed with exception {_e}')
                # Remove any steps data
                utils.get_tracked_data()
                utils.get_user_set_parameters(remove=True)
            finally:
                self.context_step_log = None  # type: ignore

            if attempts == max_attempts:
                step_log.status = defaults.FAIL
                logger.error(f'Node {node} failed, max retries of {max_attempts} reached')

        self._sync_catalog(node, step_log, stage='put', synced_catalogs=data_catalogs_get)
        self.run_log_store.add_step_log(step_log, self.run_id)

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        raise NotImplementedError

    def add_code_identities(self, node: BaseNode, step_log: StepLog, **kwargs):
        """
        Add code identities specific to the implementation.

        The Base class has an implementation of adding git code identities.

        Args:
            step_log (object): The step log object
            node (BaseNode): The node we are adding the step log for
        """
        step_log.code_identities.append(utils.get_git_code_identity(self.run_log_store))

    def execute_from_graph(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        This is the entry point to from the graph execution.

        While the self.execute_graph is responsible for traversing the graph, this function is responsible for
        actual execution of the node.

        If the node type is:
            * task : We can delegate to _execute_node after checking the eligibility for re-run in cases of a re-run
            * success: We can delegate to _execute_node
            * fail: We can delegate to _execute_node

        For nodes that are internally graphs:
            * parallel: Delegate the responsibility of execution to the node.execute_as_graph()
            * dag: Delegate the responsibility of execution to the node.execute_as_graph()
            * map: Delegate the responsibility of execution to the node.execute_as_graph()

        Check the implementations of local, local-container, local-aws-batch for different examples of implementation

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to None.
        """
        step_log = self.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING

        # Add the step log to the database as per the situation.
        # If its a terminal node, complete it now
        if node.node_type in ['success', 'fail']:
            self.run_log_store.add_step_log(step_log, self.run_id)
            self._execute_node(node, map_variable=map_variable, **kwargs)
            return

        # In single step
        if self.single_step:
            # If the node name does not match, we move on to the next node.
            if not node.name == self.single_step:
                step_log.mock = True
                step_log.status = defaults.SUCCESS
                self.run_log_store.add_step_log(step_log, self.run_id)
                return
        else:  # We are not in single step mode
            # If previous run was successful, move on to the next step
            if not self._is_eligible_for_rerun(node, map_variable=map_variable):
                step_log.mock = True
                step_log.status = defaults.SUCCESS
                self.run_log_store.add_step_log(step_log, self.run_id)
                return

        # We call an internal function to iterate the sub graphs and execute them
        if node.is_composite:
            self.run_log_store.add_step_log(step_log, self.run_id)
            node.execute_as_graph(self, map_variable=map_variable, **kwargs)
            return

        # Executor specific way to trigger a job
        self.run_log_store.add_step_log(step_log, self.run_id)
        self.trigger_job(node=node, map_variable=map_variable, **kwargs)

    def trigger_job(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        Executor specific way of triggering jobs.

        Args:
            node (BaseNode): The node to execute
            map_variable (str, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to ''.

        Raises: NotImplementedError Base class hence not implemented
        """
        raise NotImplementedError

    def _get_status_and_next_node_name(self, current_node: BaseNode, dag: Graph, map_variable: dict = None):
        """
        Given the current node and the graph, returns the name of the next node to execute.

        The name is always relative the graph that the node resides in.

        If the current node succeeded, we return the next node as per the graph.
        If the current node failed, we return the on failure node of the node (if provided) or the global one.

        Args:
            current_node (BaseNode): The current node.
            dag (Graph): The dag we are traversing.
            map_variable (dict): If the node belongs to a map branch.
        """

        step_log = self.run_log_store.get_step_log(current_node._get_step_log_name(map_variable), self.run_id)
        logger.info(
            f'Finished executing the node {current_node} with status {step_log.status}')

        next_node_name = current_node._get_next_node()

        if step_log.status == defaults.FAIL:
            next_node_name = dag.get_fail_node().name
            if current_node._get_on_failure_node():
                next_node_name = current_node._get_on_failure_node()

        return step_log.status, next_node_name

    def execute_graph(self, dag: Graph, map_variable: dict = None, **kwargs):
        """
        The parallelization is controlled by the nodes and not by this function.

        Logically the method should:
            * Start at the dag.start_at of the dag.
            * Call the self.execute_from_graph(node)
            * depending upon the status of the execution, either move to the success node or failure node.

        Args:
            dag (Graph): The directed acyclic graph to traverse and execute.
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of the iterable.
                    Defaults to None.
        """
        current_node = dag.start_at
        previous_node = None
        logger.info(f'Running the execution with {current_node}')
        while True:
            working_on = dag.get_node_by_name(current_node)

            if previous_node == current_node:
                raise Exception('Potentially running in a infinite loop')

            previous_node = current_node

            logger.info(f'Creating execution log for {working_on}')
            self.execute_from_graph(working_on, map_variable=map_variable, **kwargs)

            status, next_node_name = self._get_status_and_next_node_name(
                current_node=working_on, dag=dag, map_variable=map_variable)

            if status == defaults.TRIGGERED:
                # Some nodes go into triggered state and self traverse
                logger.info(f'Triggered the job to execute the node {current_node}')
                break

            if working_on.node_type in ['success', 'fail']:
                break

            current_node = next_node_name

        run_log = self.run_log_store.get_branch_log(working_on._get_branch_log_name(map_variable), self.run_id)

        branch = 'graph'
        if working_on.internal_branch_name:
            branch = working_on.internal_branch_name

        logger.info(f'Finished execution of the {branch} with status {run_log.status}')
        print(json.dumps(run_log.dict(), indent=4))

    def _is_eligible_for_rerun(self, node: BaseNode, map_variable: dict = None):
        """
        In case of a re-run, this method checks to see if the previous run step status to determine if a re-run is
        necessary.
            * True: If its not a re-run.
            * True: If its a re-run and we failed in the last run or the corresponding logs do not exist.
            * False: If its a re-run and we succeeded in the last run.

        Most cases, this logic need not be touched

        Args:
            node (Node): The node to check against re-run
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable..
                        Defaults to None.

        Returns:
            bool: Eligibility for re-run. True means re-run, False means skip to the next step.
        """
        if self.previous_run_log:
            node_step_log_name = node._get_step_log_name(map_variable=map_variable)
            logger.info(f'Scanning previous run logs for node logs of: {node_step_log_name}')

            previous_node_log = None
            try:
                previous_node_log, _ = self.previous_run_log.search_step_by_internal_name(node_step_log_name)
            except exceptions.StepLogNotFoundError:
                logger.warning(f'Did not find the node {node.name} in previous run log')
                return True  # We should re-run the node.

            step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
            logger.info(f'The original step status: {previous_node_log.status}')

            if previous_node_log.status == defaults.SUCCESS:
                logger.info(f'The step {node.name} is marked success, not executing it')
                step_log.status = defaults.SUCCESS
                step_log.message = 'Node execution successful in previous run, skipping it'
                self.run_log_store.add_step_log(step_log, self.run_id)
                return False  # We need not run the node

            #  Remove previous run log to start execution from this step
            logger.info(f'The new execution should start executing graph from this node {node.name}')
            self.previous_run_log = None
        return True

    def send_return_code(self, stage='traversal'):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        run_id = self.run_id

        run_log = self.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
        if run_log.status == defaults.FAIL:
            raise Exception('Pipeline execution failed')

    def _resolve_node_config(self, node: BaseNode):
        """
        The mode_config section can contain specific over-rides to an global executor config.
        To avoid too much clutter in the dag definition, we allow the configuration file to have placeholders block.
        The nodes can over-ride the global config by referring to key in the placeholder.

        For example:
        # configuration.yaml
        mode:
          type: cloud-implementation
          config:
            k1: v1
            k3: v3
            placeholders:
              k2: v2 # Could be a mapping internally.

        # in pipeline definition.yaml
        dag:
          steps:
            step1:
              mode_config:
                cloud-implementation:
                  k1: value_specific_to_node
                  k2:

        This method should resolve the node_config to {'k1': 'value_specific_to_node', 'k2': 'v2', 'k3': 'v3'}

        Args:
            node (BaseNode): The current node being processed.

        """
        effective_node_config = copy.deepcopy(self.config.dict())
        ctx_node_config = node._get_mode_config(self.service_name)

        placeholders = self.config.placeholders

        for key, value in ctx_node_config.items():
            if not value:
                if key in placeholders:  # Update via placeholder only if value is None
                    try:
                        effective_node_config.update(placeholders[key])
                    except TypeError:
                        logger.error(f'Expected value to the {key} to be a mapping but found {type(placeholders[key])}')
                    continue
                logger.info(f"For key: {key} in the {node.name} mode_config, there is no value provided and no \
                    corresponding placeholder was found")

            effective_node_config[key] = value

        effective_node_config.pop('placeholders', None)

        return effective_node_config


class LocalExecutor(BaseExecutor):
    """
    In the mode of local execution, we run everything on the local computer.

    This has some serious implications on the amount of time it would take to complete the run.
    Also ensure that the local compute is good enough for the compute to happen of all the steps.

    Example config:
    mode:
      type: local
      config:
        enable_parallel: True or False to enable parallel.

    """
    service_name = 'local'

    def trigger_job(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        In this mode of execution, we prepare for the node execution and execute the node

        Args:
            node (BaseNode): [description]
            map_variable (str, optional): [description]. Defaults to ''.
        """
        self.prepare_for_node_execution()
        self.execute_node(node=node, map_variable=map_variable, **kwargs)

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        self._execute_node(node=node, map_variable=map_variable, **kwargs)


class LocalContainerExecutor(BaseExecutor):
    """
    In the mode of local-container, we execute all the commands in a container.

    Ensure that the local compute has enough resources to finish all your jobs.

    The image of the run, could either be provided as default in the configuration of the mode
    i.e.:
    mode:
      type: 'local-container'
      config:
        docker_image: the image you want the code to run in.

    or default image could be over-ridden for a single node by providing a docker_image in the step config.
    i.e:
    dag:
      steps:
        step:
          mode_config:
            local-container:
                docker_image: The image that you want that single step to run in.
    This image would only be used for that step only.

    This mode does not build the docker image with the latest code for you, it is still left for the user to build
    and ensure that the docker image provided is the correct one.

    Example config:
    mode:
      type: local-container
      config:
        docker_image: The default docker image to use if the node does not provide one.
    """
    service_name = 'local-container'

    class Config(BaseExecutor.Config):
        docker_image: str

    def __init__(self, config):
        # pylint: disable=R0914,R0913
        super().__init__(config=config)
        self.container_log_location = '/tmp/run_logs/'
        self.container_catalog_location = '/tmp/catalog/'
        self.container_secrets_location = '/tmp/dotenv'
        self.volumes = {}

    @ property
    def docker_image(self):
        return self.config.docker_image

    def add_code_identities(self, node: BaseNode, step_log: StepLog, **kwargs):
        """
        Call the Base class to add the git code identity and add docker identity

        Args:
            node (BaseNode): The node we are adding the code identity
            step_log (Object): The step log corresponding to the node
        """

        super().add_code_identities(node, step_log)
        mode_config = self._resolve_node_config(node)

        docker_image = mode_config.get('docker_image', None)
        if docker_image:
            code_id = self.run_log_store.create_code_identity()

            code_id.code_identifier = utils.get_local_docker_image_id(docker_image)
            code_id.code_identifier_type = 'docker'
            code_id.code_identifier_dependable = True
            code_id.code_identifier_url = 'local docker host'
            step_log.code_identities.append(code_id)

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        return self._execute_node(node, map_variable, **kwargs)

    def trigger_job(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        If the config has "run_in_local: True", we compute it on local system instead of container.
        In local container mode, we just spin the container to execute magnus execute_single_node.

        Args:
            node (BaseNode): The node we are currently executing
            map_variable (str, optional): If the node is part of the map branch. Defaults to ''.
        """
        mode_config = self._resolve_node_config(node)

        if "run_in_local" in mode_config and mode_config["run_in_local"]:
            # Do not change config but only validate the configuration.
            integration.validate(self, self.run_log_store)
            integration.validate(self, self.catalog_handler)
            integration.validate(self, self.secrets_handler)

            self.execute_node(node=node, map_variable=map_variable, **kwargs)
            return

        self._spin_container(node, map_variable=map_variable, **kwargs)

        # Check for the status of the node log and anything apart from Success is FAIL
        # This typically happens if something is wrong with magnus or settings.
        step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
        if step_log.status != defaults.SUCCESS:
            msg = (
                'Node execution inside the container failed. Please check the logs.\n'
                'Note: If you do not see any docker issue from your side and the code works properly on local mode '
                'please raise a bug report.'
            )
            logger.warning(msg)
            step_log.status = defaults.FAIL
            self.run_log_store.add_step_log(step_log, self.run_id)

    def _spin_container(self, node, map_variable: dict = None, **kwargs):  # pylint: disable=unused-argument
        """
        During the flow run, we have to spin up a container with the docker image mentioned
        and the right log locations
        """
        # Conditional import
        import docker  # pylint: disable=C0415

        try:
            client = docker.from_env()
        except Exception as ex:
            logger.exception('Could not get access to docker')
            raise Exception('Could not get the docker socket file, do you have docker installed?') from ex

        try:
            action = utils.get_node_execution_command(self, node, map_variable=map_variable)
            logger.info(f'Running the command {action}')
            #  Overrides global config with local
            mode_config = self._resolve_node_config(node)
            docker_image = mode_config.get('docker_image', None)
            environment = mode_config.get('environment', {})
            environment.update(self.variables)
            if not docker_image:
                raise Exception(
                    f'Please provide a docker_image using mode_config of the step {node.name} or at global mode')

            # TODO: Should consider using getpass.getuser() when running the docker container? Volume permissions
            container = client.containers.create(image=docker_image,
                                                 command=action,
                                                 auto_remove=True,
                                                 volumes=self.volumes,
                                                 network_mode='host',
                                                 environment=environment)
            container.start()
            stream = container.logs(stream=True, follow=True)
            while True:
                try:
                    output = next(stream).decode("utf-8")
                    output = output.strip('\r\n')
                    logger.info(output)
                except StopIteration:
                    logger.info('Docker Run completed')
                    break

        except Exception as _e:
            logger.exception('Problems with spinning up the container')
            raise _e


class DemoRenderer(BaseExecutor):
    """
    This renderer is an example of how you can render required job specifications as per your orchestration tool.

    BaseExecutor implements many of the functionalities that are common and can be safe defaults.
    In this renderer example: We just render a bash script that sequentially calls the steps.
    We do not handle composite steps in this mode.

    Example config:
    mode:
      type: demo-renderer
    """
    service_name = 'demo-renderer'

    def execute_node(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        This method does the actual execution of a task, as-is, success or fail node.
        """
        self._set_up_run_log(exists_ok=True)
        # Need to set up the step log for the node as the entry point is different
        step_log = self.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self.run_log_store.add_step_log(step_log, self.run_id)

        super()._execute_node(node, map_variable=map_variable, **kwargs)

        step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
        if step_log.status == defaults.FAIL:
            raise Exception(f'Step {node.name} failed')

    def trigger_job(self, node: BaseNode, map_variable: dict = None, **kwargs):
        """
        Executor specific way of triggering jobs.

        This method has to be changed to do what exactly you want as part of your computational engine

        If your compute is not local, use utils.get_node_execution_command(self, node, map_variable=map_variable)
        to get the command to run a single node.

        If the compute is local to the environment, calls prepare_for_node_execution and call _execute_node
        NOTE: This method should always be implemented.
        """
        self.prepare_for_node_execution()
        self.execute_node(node=node, map_variable=map_variable, **kwargs)

    def send_return_code(self, stage='traversal'):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        if stage != 'traversal':  # traversal does no actual execution, so return code is pointless
            run_id = self.run_id

            run_log = self.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
            if run_log.status == defaults.FAIL:
                raise Exception('Pipeline execution failed')

    def execute_graph(self, dag: Graph, map_variable: dict = None, **kwargs):
        """
        Iterate through the graph and frame the bash script.

        For more complex outputs, dataclasses might be a better option.

        NOTE: This method should be over-written to write the exact specification to the compute engine.

        """
        current_node = dag.start_at
        previous_node = None
        logger.info(f'Rendering job started at {current_node}')
        bash_script_lines = []

        while True:
            working_on = dag.get_node_by_name(current_node)

            if working_on.is_composite:
                raise NotImplementedError('In this demo version, composite nodes are not implemented')

            if previous_node == current_node:
                raise Exception('Potentially running in a infinite loop')

            previous_node = current_node

            logger.info(f'Creating execution log for {working_on}')

            _execute_node_command = utils.get_node_execution_command(self, working_on, over_write_run_id='$1')
            current_job_id = re.sub('[^A-Za-z0-9]+', '', f'{current_node}_job_id')
            fail_node_command = utils.get_node_execution_command(self, dag.get_fail_node(), over_write_run_id='$1')

            if working_on.node_type not in ['success', 'fail']:
                if working_on.node_type == 'as-is':
                    command_config = working_on.config.get('command_config', {})
                    if 'render_string' in command_config:
                        bash_script_lines.append(command_config['render_string'] + '\n')
                else:
                    bash_script_lines.append(f'{_execute_node_command}\n')

                bash_script_lines.append('exit_code=$?\necho $exit_code\n')
                # Write failure node
                bash_script_lines.append(
                    (
                        'if [ $exit_code -ne 0 ];\nthen\n'
                        f'\t $({fail_node_command})\n'
                        '\texit 1\n'
                        'fi\n'
                    )
                )

            if working_on.node_type == 'success':
                bash_script_lines.append(f'{_execute_node_command}')
            if working_on.node_type in ['success', 'fail']:
                break

            current_node = working_on._get_next_node()

        with open('demo-bash.sh', 'w', encoding='utf-8') as fw:
            fw.writelines(bash_script_lines)

        msg = (
            'demo-bash.sh for running the pipeline is written. To execute it \n'
            '1). Activate the environment:\n'
            '\t for example poetry shell or pipenv shell etc\n'
            '2). Make the shell script executable.\n'
            '\t chmod 755 demo-bash.sh\n'
            '3). Run the script by: source demo-bash.sh <run_id>\n'
            '\t The first argument to the script is the run id you want for the run.'
        )
        logger.info(msg)



import logging
from logging.config import fileConfig

import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points, resource_filename

from magnus import defaults, docker_utils, pipeline

fileConfig(resource_filename(__name__, 'log_config.ini'))
logger = logging.getLogger(defaults.NAME)


@with_plugins(iter_entry_points('magnus.cli_plugins'))
@click.group()
@click.version_option()
def cli():
    """
    Welcome to magnus. Please provide the command that you want to use.
    All commands have options that you can see by magnus <command> --help
    """
    pass


@cli.command('execute', short_help="Execute/translate a pipeline")
@click.option('-f', '--file', default='pipeline.yaml', help='The pipeline definition file', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--tag', help='A tag attached to the run')
@click.option('--run-id', help='An optional run_id, one would be generated if not provided')
@click.option('--use-cached', help='Provide the previous run_id to re-run.', show_default=True)
def execute(file, config_file, parameters_file, log_level, tag, run_id, use_cached):  # pragma: no cover
    """
    Entry point to executing a pipeline. This command is most commonly used
    either to execute a pipeline or to translate the pipeline definition to another language.

    You can re-run an older run by providing the run_id of the older run in --use-cached.
    Ensure that the catalogs and run logs are accessible by the present configuration.
    """
    logger.setLevel(log_level)
    pipeline.execute(
        configuration_file=config_file, pipeline_file=file, tag=tag,
        run_id=run_id, use_cached=use_cached, parameters_file=parameters_file)


@cli.command('execute_step', short_help="Execute a single step of the pipeline")
@click.argument('step_name')
@click.option('-f', '--file', default='pipeline.yaml', help='The pipeline definition file', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--tag', help='A tag attached to the run')
@click.option('--run-id', help='An optional run_id, one would be generated if not provided')
@click.option('--use-cached', help='Provide the previous run_id to re-run.', show_default=True)
def execute_step(step_name, file, config_file, parameters_file, log_level, tag, run_id, use_cached):  # pragma: no cover
    """
    Entry point to executing a single step of the pipeline.

    This command is helpful to run only one step of the pipeline in isolation.
    Only the steps of the parent dag could be invoked using this method.

    You can re-run an older run by providing the run_id of the older run in --use-cached.
    Ensure that the catalogs and run logs are accessible by the present configuration.

    When running map states, ensure that the parameter to iterate on is available in parameter space.
    """
    logger.setLevel(log_level)
    pipeline.execute_single_step(
        configuration_file=config_file, pipeline_file=file, step_name=step_name, tag=tag,
        run_id=run_id, parameters_file=parameters_file, use_cached=use_cached)


@cli.command("execute_single_node", short_help="Internal entry point to execute a single node", hidden=True)
@click.argument('run_id')
@click.argument('step_name')
@click.option('--map-variable', default='', help='The map variable dictionary in str', show_default=True)
@click.option('-f', '--file', default='', help='The pipeline definition file', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--tag', default='', help='A tag attached to the run')
def execute_single_node(run_id, step_name, map_variable, file, config_file, parameters_file, log_level, tag):
    logger.setLevel(log_level)

    pipeline.execute_single_node(configuration_file=config_file, pipeline_file=file,
                                 step_name=step_name, map_variable=map_variable,
                                 run_id=run_id, tag=tag, parameters_file=parameters_file)


@cli.command("execute_single_branch", short_help="Internal entry point to execute a single branch", hidden=True)
@click.argument('run_id')
@click.argument('branch_name')
@click.option('--map-variable', default='', help='The map variable dictionary in str', show_default=True)
@click.option('-f', '--file', default='pipeline.yaml', help='The pipeline definition file', show_default=True)
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
def execute_single_branch(run_id, branch_name, map_variable, file, config_file, log_level):
    logger.setLevel(log_level)

    pipeline.execute_single_brach(configuration_file=config_file,
                                  pipeline_file=file, branch_name=branch_name, map_variable=map_variable,
                                  run_id=run_id)


@cli.command("execute_notebook", short_help="Entry point to execute a notebook")
@click.argument('filename')
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--data-folder', '-d', default='data/', help="The catalog data folder")
@click.option('--put-in-catalog', '-put', default=None, multiple=True, help="The data to put from the catalog")
@click.option('--tag', help='A tag attached to the run')
@click.option('--run-id', help='An optional run_id, one would be generated if not provided')
def execute_notebook(filename, config_file, parameters_file, log_level, data_folder,
                     put_in_catalog, tag, run_id):
    """
    Entry point to execute a Jupyter notebook in isolation.

    The notebook would be executed in the environment defined by the config file or default if none.
    """
    logger.setLevel(log_level)
    catalog_config = {
        'compute_data_folder': data_folder,
        'put': list(put_in_catalog) if put_in_catalog else None
    }
    pipeline.execute_notebook(notebook_file=filename, catalog_config=catalog_config, configuration_file=config_file,
                              parameters_file=parameters_file, tag=tag, run_id=run_id)


@cli.command("execute_function", short_help="Entry point to execute a python function")
@click.argument('command')
@click.option('-c', '--config-file', default=None,
              help="config file, in yaml, to be used for the run", show_default=True)
@click.option('-p', '--parameters-file', default=None,
              help="Parameters, in yaml,  accessible by the application", show_default=True)
@click.option('--log-level', default=defaults.LOG_LEVEL, help='The log level', show_default=True,
              type=click.Choice(['INFO', 'DEBUG', 'WARNING', 'ERROR', 'FATAL']))
@click.option('--data-folder', '-d', default='data/', help="The catalog data folder")
@click.option('--put-in-catalog', '-put', default=None, multiple=True, help="The data to put from the catalog")
@click.option('--tag', help='A tag attached to the run')
@click.option('--run-id', help='An optional run_id, one would be generated if not provided')
def execute_function(command, config_file, parameters_file, log_level, data_folder,
                     put_in_catalog, tag, run_id):
    """
    Entry point to execute a python function in isolation.

    The function would be executed in the environment defined by the config file or default if none.
    """
    logger.setLevel(log_level)
    catalog_config = {
        'compute_data_folder': data_folder,
        'put': list(put_in_catalog) if put_in_catalog else None
    }
    pipeline.execute_function(command=command, catalog_config=catalog_config, configuration_file=config_file,
                              parameters_file=parameters_file, tag=tag, run_id=run_id)


@cli.command('build_docker', short_help="Utility tool to build docker images")
@click.argument('image_name')
@click.option('-f', '--docker-file', default=None, help="The dockerfile to be used. If None, we generate one")
@click.option('-s', '--style', default="poetry", help="The method used to get requirements", show_default=True)
@click.option('-t', "--tag", default="latest", help="The tag assigned to the image", show_default=True)
@click.option('-c', '--commit-tag', is_flag=True, default=False, help="Use commit id as tag. Over-rides tag option",
              show_default=True)
@click.option('-d', '--dry-run', is_flag=True, default=False, help="Generate the dockerfile, but NOT the image",
              show_default=True)
@click.option('--git-tracked/--all', default=True, help="Controls what should be added to image. All vs git-tracked",
              show_default=True)
def build_docker(image_name, docker_file, style, tag, commit_tag, dry_run, git_tracked):
    """
    A utility function to create docker images from the existing codebase.

    It is advised to provide your own dockerfile as much as possible. If you do not have one handy, you can use
    --dry-run functionality to see if the auto-generated one suits your needs.

    If you are auto-generating the dockerfile:
    BEWARE!! Over-riding the default options assumes you know what you are doing! BEWARE!!

    1). By default, only git tracked files are added to the docker image.

    2). The auto-generated dockerfile uses, python 3.8 as the default image and adds the current folder.
    """
    docker_utils.build_docker(image_name=image_name, docker_file=docker_file, style=style,
                              tag=tag, commit_tag=commit_tag, dry_run=dry_run, git_tracked=git_tracked)



import json
import logging
from typing import Optional, Union

from magnus import defaults, exceptions, graph, utils

logger = logging.getLogger(defaults.NAME)

# Set this global executor to the fitted executor for access later
global_executor = None  # pylint: disable=invalid-name # type: ignore

# TODO: Tests and mypy


def get_default_configs() -> dict:
    """
    User can provide extensions as part of their code base, magnus-config.yaml provides the place to put them.
    """
    user_configs = {}
    if utils.does_file_exist(defaults.USER_CONFIG_FILE):
        user_configs = utils.load_yaml(defaults.USER_CONFIG_FILE)

    if not user_configs:
        return {}

    user_defaults = user_configs.get('defaults', {})
    if user_defaults:
        return user_defaults

    return {}


def prepare_configurations(
        configuration_file: str = None,
        pipeline_file: str = None,
        run_id: str = None,
        tag: Union[str, None] = None,
        use_cached: Union[str, None] = '',
        parameters_file: str = None):
    # pylint: disable=R0914
    """
    Replace the placeholders in the dag/config against the variables file.

    Attach the secrets_handler, run_log_store, catalog_handler to the executor and return it.

    Args:
        variables_file (str): The variables file, if used or None
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        use_cached (str): Provide the run_id of the older run

    Returns:
        executor.BaseExecutor : A prepared executor as per the dag/config
    """
    magnus_defaults = get_default_configs()

    variables = utils.gather_variables()

    configuration = {}
    if configuration_file:
        configuration = utils.load_yaml(configuration_file)

    # apply variables
    configuration = utils.apply_variables(configuration, variables)

    # Run log settings, configuration over-rides everything
    run_log_config = configuration.get('run_log_store', {})
    if not run_log_config:
        run_log_config = magnus_defaults.get('run_log_store', defaults.DEFAULT_RUN_LOG_STORE)
    run_log_store = utils.get_provider_by_name_and_type('run_log_store', run_log_config)

    # Catalog handler settings, configuration over-rides everything
    catalog_config = configuration.get('catalog', {})
    if not catalog_config:
        catalog_config = magnus_defaults.get('catalog', defaults.DEFAULT_CATALOG)
    catalog_handler = utils.get_provider_by_name_and_type('catalog', catalog_config)

    # Secret handler settings, configuration over-rides everything
    secrets_config = configuration.get('secrets', {})
    if not secrets_config:
        secrets_config = magnus_defaults.get('secrets', defaults.DEFAULT_SECRETS)
    secrets_handler = utils.get_provider_by_name_and_type('secrets', secrets_config)

    # experiment tracker settings, configuration over-rides everything
    tracker_config = configuration.get('experiment_tracking', {})
    if not tracker_config:
        tracker_config = magnus_defaults.get('experiment_tracking', defaults.DEFAULT_EXPERIMENT_TRACKER)
    tracker_handler = utils.get_provider_by_name_and_type('experiment_tracking', tracker_config)

    # Mode configurations, configuration over rides everything
    mode_config = configuration.get('mode', {})
    if not mode_config:
        mode_config = magnus_defaults.get('executor', defaults.DEFAULT_EXECUTOR)
    mode_executor = utils.get_provider_by_name_and_type('executor', mode_config)

    if pipeline_file:
        # There are use cases where we are only preparing the executor
        pipeline_config = utils.load_yaml(pipeline_file)
        pipeline_config = utils.apply_variables(pipeline_config, variables=variables)

        logger.info('The input pipeline:')
        logger.info(json.dumps(pipeline_config, indent=4))

        # Create the graph
        dag_config = pipeline_config['dag']
        dag_hash = utils.get_dag_hash(dag_config)
        # TODO: Dag nodes should not self refer themselves
        dag = graph.create_graph(dag_config)

        mode_executor.pipeline_file = pipeline_file
        mode_executor.dag = dag
        mode_executor.dag_hash = dag_hash

    mode_executor.run_id = run_id
    mode_executor.tag = tag
    mode_executor.use_cached = use_cached

    # Set a global executor for inter-module access later
    global global_executor  # pylint: disable=W0603,invalid-name,
    global_executor = mode_executor

    mode_executor.run_log_store = run_log_store
    mode_executor.catalog_handler = catalog_handler
    mode_executor.secrets_handler = secrets_handler
    mode_executor.experiment_tracker = tracker_handler
    mode_executor.configuration_file = configuration_file
    mode_executor.parameters_file = parameters_file
    mode_executor.variables = variables

    return mode_executor


def execute(
        configuration_file: str,
        pipeline_file: str,
        tag: str = None,
        run_id: str = None,
        use_cached: str = None,
        parameters_file: str = None):
    # pylint: disable=R0914,R0913
    """
    The entry point to magnus execution. This method would prepare the configurations and delegates traversal to the
    executor

    Args:
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        use_cached (str): The previous run_id to use.
        parameters_file (str): The parameters being sent in to the application
    """
    # Re run settings
    run_id = utils.generate_run_id(run_id=run_id)

    mode_executor = prepare_configurations(configuration_file=configuration_file,
                                           pipeline_file=pipeline_file,
                                           run_id=run_id,
                                           tag=tag,
                                           use_cached=use_cached,
                                           parameters_file=parameters_file)
    mode_executor.execution_plan = defaults.EXECUTION_PLAN.pipeline

    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    previous_run_log = None
    if use_cached:
        try:
            previous_run_log = mode_executor.run_log_store.get_run_log_by_id(run_id=use_cached, full=True)
        except exceptions.RunLogNotFoundError as _e:
            msg = (
                f'There is no run by {use_cached} in the current run log store '
                f'{mode_executor.run_log_store.service_name}. Please ensure that that run log exists to re-run.\n'
                'Note: Even if the previous run used a different run log store, provide the run log store in the format'
                ' accepted by the current run log store.'
            )
            raise Exception(msg) from _e

        if previous_run_log.dag_hash != mode_executor.dag_hash:
            logger.warning('The previous dag does not match to the current one!')
        mode_executor.previous_run_log = previous_run_log
        logger.info('Found a previous run log and using it as cache')

    # Prepare for graph execution
    mode_executor.prepare_for_graph_execution()

    logger.info('Executing the graph')
    mode_executor.execute_graph(dag=mode_executor.dag)

    mode_executor.send_return_code()


def execute_single_step(
        configuration_file: str,
        pipeline_file: str,
        step_name: str,
        run_id: str,
        tag: str = None,
        parameters_file: str = None,
        use_cached: str = None,):
    # pylint: disable=R0914,R0913
    """
    The entry point into executing a single step of magnus.

    It should have similar set up of configurations to execute because orchestrator modes can initiate the execution.

    Args:
        variables_file (str): The variables file, if used or None
        step_name : The name of the step to execute in dot path convention
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        parameters_file (str): The parameters being sent in to the application

    """
    run_id = utils.generate_run_id(run_id=run_id)

    mode_executor = prepare_configurations(configuration_file=configuration_file,
                                           pipeline_file=pipeline_file,
                                           run_id=run_id,
                                           tag=tag,
                                           use_cached='',
                                           parameters_file=parameters_file)
    mode_executor.execution_plan = defaults.EXECUTION_PLAN.pipeline
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)
    try:
        _ = mode_executor.dag.get_node_by_name(step_name)
    except exceptions.NodeNotFoundError as e:
        msg = (
            f"The node by name {step_name} is not found in the graph. Please provide a valid node name"
        )
        raise Exception(msg) from e

    previous_run_log = None
    if use_cached:
        try:
            previous_run_log = mode_executor.run_log_store.get_run_log_by_id(run_id=use_cached, full=True)
        except exceptions.RunLogNotFoundError as _e:
            msg = (
                f'There is no run by {use_cached} in the current run log store '
                f'{mode_executor.run_log_store.service_name}. Please ensure that that run log exists to re-run.\n'
                'Note: Even if the previous run used a different run log store, provide the run log store in the format'
                ' accepted by the current run log store.'
            )
            raise Exception(msg) from _e

        if previous_run_log.dag_hash != mode_executor.dag_hash:
            logger.warning('The previous dag does not match to the current one!')
        mode_executor.previous_run_log = previous_run_log
        logger.info('Found a previous run log and using it as cache')

    mode_executor.single_step = step_name
    mode_executor.prepare_for_graph_execution()

    logger.info('Executing the graph')
    mode_executor.execute_graph(dag=mode_executor.dag)

    mode_executor.send_return_code()


def execute_single_node(
        configuration_file: str,
        pipeline_file: str,
        step_name: str,
        map_variable: str,
        run_id: str,
        tag: str = None,
        parameters_file: str = None):
    # pylint: disable=R0914,R0913
    """
    The entry point into executing a single node of magnus. Orchestration modes should extensively use this
    entry point.

    It should have similar set up of configurations to execute because orchestrator modes can initiate the execution.

    Args:
        variables_file (str): The variables file, if used or None
        step_name : The name of the step to execute in dot path convention
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        parameters_file (str): The parameters being sent in to the application

    """
    from magnus import nodes
    mode_executor = prepare_configurations(configuration_file=configuration_file,
                                           pipeline_file=pipeline_file,
                                           run_id=run_id,
                                           tag=tag,
                                           use_cached='',
                                           parameters_file=parameters_file)
    mode_executor.execution_plan = defaults.EXECUTION_PLAN.pipeline
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    mode_executor.prepare_for_node_execution()

    if not mode_executor.dag:
        # There are a few entry points that make graph dynamically and do not have a dag defined statically.
        run_log = mode_executor.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
        mode_executor.dag = graph.create_graph(run_log.run_config['pipeline'])

    step_internal_name = nodes.BaseNode._get_internal_name_from_command_name(step_name)

    map_variable_dict = utils.json_to_ordered_dict(map_variable)

    node_to_execute, _ = graph.search_node_by_internal_name(mode_executor.dag, step_internal_name)

    logger.info('Executing the single node of : %s', node_to_execute)
    mode_executor.execute_node(node=node_to_execute, map_variable=map_variable_dict)

    mode_executor.send_return_code(stage='execution')


def execute_single_brach(
        configuration_file: str,
        pipeline_file: str,
        branch_name: str,
        map_variable: str,
        run_id: str,
        tag: str = None):
    # pylint: disable=R0914,R0913
    """
    The entry point into executing a branch of the graph. Interactive modes in parallel runs use this to execute
    branches in parallel.

    This entry point is never used by its own but rather from a node. So the arguments sent into this are fewer.

    Args:
        variables_file (str): The variables file, if used or None
        branch_name : The name of the branch to execute, in dot.path.convention
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
    """
    from magnus import nodes
    mode_executor = prepare_configurations(configuration_file=configuration_file,
                                           pipeline_file=pipeline_file,
                                           run_id=run_id,
                                           tag=tag,
                                           use_cached='')
    mode_executor.execution_plan = defaults.EXECUTION_PLAN.pipeline
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    branch_internal_name = nodes.BaseNode._get_internal_name_from_command_name(branch_name)

    map_variable_dict = utils.json_to_ordered_dict(map_variable)

    branch_to_execute = graph.search_branch_by_internal_name(mode_executor.dag, branch_internal_name)

    logger.info('Executing the single branch of %s', branch_to_execute)
    mode_executor.execute_graph(dag=branch_to_execute, map_variable=map_variable_dict)

    mode_executor.send_return_code()


def execute_notebook(
        notebook_file: str,
        catalog_config: dict,
        configuration_file: str,
        tag: str = None,
        run_id: str = None,
        parameters_file: str = None):
    # pylint: disable=R0914,R0913
    """
    The entry point to magnus execution of a notebook. This method would prepare the configurations and
    delegates traversal to the executor
    """
    run_id = utils.generate_run_id(run_id=run_id)

    mode_executor = prepare_configurations(
        configuration_file=configuration_file,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file)

    mode_executor.execution_plan = defaults.EXECUTION_PLAN.notebook
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    # Prepare the graph with a single node
    dag = graph.Graph(start_at='executing notebook')
    step_config = {
        'command': notebook_file,
        'command_type': 'notebook',
        'type': 'task',
        'next': 'success',
        'catalog': catalog_config,
    }
    node = graph.create_node(name=f'executing notebook', step_config=step_config)

    dag.add_node(node)
    dag.add_terminal_nodes()

    mode_executor.dag = dag
    # Prepare for graph execution
    mode_executor.prepare_for_graph_execution()

    logger.info('Executing the graph')
    mode_executor.execute_graph(dag=mode_executor.dag)

    mode_executor.send_return_code()


def execute_function(
        command: str,
        catalog_config: dict,
        configuration_file: str,
        tag: str = None,
        run_id: str = None,
        parameters_file: str = None):
    # pylint: disable=R0914,R0913
    """
    The entry point to magnus execution of a function. This method would prepare the configurations and
    delegates traversal to the executor
    """
    run_id = utils.generate_run_id(run_id=run_id)

    mode_executor = prepare_configurations(
        configuration_file=configuration_file,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file)

    mode_executor.execution_plan = defaults.EXECUTION_PLAN.function
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    # Prepare the graph with a single node
    dag = graph.Graph(start_at='executing function')
    step_config = {
        'command': command,
        'command_type': 'python',
        'type': 'task',
        'next': 'success',
        'catalog': catalog_config,
    }
    node = graph.create_node(name=f'executing function', step_config=step_config)

    dag.add_node(node)
    dag.add_terminal_nodes()

    mode_executor.dag = dag
    # Prepare for graph execution
    mode_executor.prepare_for_graph_execution()

    logger.info('Executing the graph')
    mode_executor.execute_graph(dag=mode_executor.dag)

    mode_executor.send_return_code()



from __future__ import annotations

import json
import logging
from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional, OrderedDict, Tuple, Union

from pydantic import BaseModel

from magnus import defaults, exceptions, utils

logger = logging.getLogger(defaults.NAME)

# Once defined these classes are sealed to any additions unless a default is provided
# Breaking this rule might make magnus backwardly incompatible


class DataCatalog(BaseModel):
    """
    The captured attributes of a catalog item.
    """
    name: str  #  The name of the dataset
    data_hash: str = ''  # The sha1 hash of the file
    catalog_relative_path: str = ''  # The file path relative the catalog location
    catalog_handler_location: str = ''  # The location of the catalog
    stage: str = ''  # The stage at which we recorded it get, put etc

    # Needed for set operations to work on DataCatalog objects
    def __hash__(self):
        return hash(self.name)

    # Needed for set operations to work on DataCatalog objects
    def __eq__(self, other):
        if not isinstance(other, DataCatalog):
            return False
        return other.name == self.name


class StepAttempt(BaseModel):
    """
    The captured attributes of an Attempt of a step.
    """
    attempt_number: int = 0
    start_time: str = ''
    end_time: str = ''
    duration: str = ''  #  end_time - start_time
    status: str = 'FAIL'
    message: str = ''


class CodeIdentity(BaseModel):
    """
    The captured attributes of a code identity of a step.
    """
    code_identifier: Optional[str] = ''  # GIT sha code or docker image id
    code_identifier_type: Optional[str] = ''  # git or docker
    code_identifier_dependable: Optional[bool] = False  # If git, checks if the tree is clean.
    code_identifier_url: Optional[str] = ''  # The git remote url or docker repository url
    code_identifier_message: Optional[str] = ''  # Any optional message


class StepLog(BaseModel):
    """
    The data class capturing the data of a Step
    """
    name: str
    internal_name: str  # Should be the dot notation of the step
    status: str = 'FAIL'  #  Should have a better default
    step_type: str = 'task'
    message: str = ''
    mock: bool = False
    code_identities: List[CodeIdentity] = []
    attempts: List[StepAttempt] = []
    user_defined_metrics: dict = {}
    branches: Dict[str, BranchLog] = {}  # Keyed in by the branch key name
    data_catalog: List[DataCatalog] = []

    def get_data_catalogs_by_stage(self, stage='put') -> List[DataCatalog]:
        """
        Given a stage, return the data catalogs according to the stage

        Args:
            stage (str, optional): The stage at which the data was cataloged. Defaults to 'put'.

        Raises:
            Exception: If the stage was not in get or put.

        Returns:
            List[DataCatalog]: The list of data catalogs as per the stage.
        """
        if stage not in ['get', 'put']:
            raise Exception('Stage should be in get or put')

        data_catalogs = []
        if self.branches:
            for _, branch in self.branches.items():
                data_catalogs.extend(branch.get_data_catalogs_by_stage(stage=stage))

        return [dc for dc in self.data_catalog if dc.stage == stage] + data_catalogs

    def add_data_catalogs(self, data_catalogs: List[DataCatalog]):
        """
        Add the data catalogs as asked by the user

        Args:
            dict_catalogs ([DataCatalog]): A list of data catalog items
        """
        if not self.data_catalog:
            self.data_catalog = []
        for data_catalog in data_catalogs:
            self.data_catalog.append(data_catalog)


class BranchLog(BaseModel):
    """
    The dataclass of captured data about a branch of a composite node.

    Returns:
        [type]: [description]
    """
    internal_name: str
    status: str = 'FAIL'
    steps: OrderedDict[str, StepLog] = {}  # type: ignore # StepLogs keyed by internal name

    def get_data_catalogs_by_stage(self, stage='put') -> List[DataCatalog]:
        """
        Given a stage, return the data catalogs according to the stage

        Args:
            stage (str, optional): The stage at which the data was cataloged. Defaults to 'put'.

        Raises:
            Exception: If the stage was not in get or put.

        Returns:
            List[DataCatalog]: The list of data catalogs as per the stage.
        """
        if stage not in ['get', 'put']:
            raise Exception('Stage should be in get or put')

        data_catalogs = []
        for _, step in self.steps.items():
            data_catalogs.extend(step.get_data_catalogs_by_stage(stage=stage))

        return data_catalogs


# Needed for BranchLog of StepLog to be referenced
StepLog.update_forward_refs()


class RunLog(BaseModel):
    """
    The data captured as part of Run Log
    """
    run_id: str
    dag_hash: Optional[str] = None
    use_cached: bool = False
    tag: Optional[str] = ''
    original_run_id: Optional[str] = ''
    status: str = defaults.FAIL
    steps: OrderedDict[str, StepLog] = {}  # type: ignore # Has the steps keyed by internal_name
    parameters: dict = {}
    run_config: dict = {}

    def get_data_catalogs_by_stage(self, stage: str = 'put') -> List[DataCatalog]:
        """
        Return all the cataloged data by the stage at which they were cataloged.

        Raises:
            Exception: If stage was not either put or get.

        Args:
            stage (str, optional): [description]. Defaults to 'put'.
        """
        if stage not in ['get', 'put']:
            raise Exception('Only get or put are allowed in stage')

        data_catalogs = []
        for _, step in self.steps.items():
            data_catalogs.extend(step.get_data_catalogs_by_stage(stage=stage))

        return list(set(data_catalogs))

    def search_branch_by_internal_name(self, i_name: str) -> Tuple[Union[BranchLog, RunLog], Union[StepLog, None]]:
        """
        Given a branch internal name, search for it in the run log.

        If the branch internal name is none, its the run log itself.

        Args:
            i_name (str): [description]

        Raises:
            exceptions.BranchLogNotFoundError: [description]

        Returns:
            Tuple[BranchLog, StepLog]: [description]
        """
        # internal name is null for base dag
        if not i_name:
            return self, None

        dot_path = i_name.split('.')

        # any internal name of a branch when split against .
        # goes step.branch.step.branch
        # If its odd, its a step, if its even its a branch
        current_steps = self.steps
        current_step = None
        current_branch = None

        for i in range(len(dot_path)):
            if i % 2:
                # Its odd, so we are in branch
                # Get the branch that holds the step
                current_branch = current_step.branches['.'.join(dot_path[:i + 1])]  # type: ignore
                current_steps = current_branch.steps
                logger.debug(f'Finding branch {i_name} in branch: {current_branch}')
            else:
                # Its even, so we are in step, we start here!
                # Get the step that holds the branch
                current_step = current_steps['.'.join(dot_path[:i + 1])]
                logger.debug(f'Finding branch for {i_name} in step: {current_step}')

        logger.debug(f'current branch : {current_branch}, current step {current_step}')
        if current_branch and current_step:
            return current_branch, current_step

        raise exceptions.BranchLogNotFoundError(self.run_id, i_name)

    def search_step_by_internal_name(self, i_name: str) -> Tuple[StepLog, Union[BranchLog, None]]:
        """
        Given a steps internal name, search for the step name.

        If the step name when split against '.' is 1, it is the run log

        Args:
            i_name (str): [description]

        Raises:
            exceptions.StepLogNotFoundError: [description]

        Returns:
            Tuple[StepLog, BranchLog]: [description]
        """
        dot_path = i_name.split('.')
        if len(dot_path) == 1:
            return self.steps[i_name], None

        current_steps = self.steps
        current_step = None
        current_branch = None
        for i in range(len(dot_path)):
            if i % 2:
                # Its odd, so we are in brach name
                current_branch = current_step.branches['.'.join(dot_path[:i + 1])]  # type: ignore
                current_steps = current_branch.steps
                logger.debug(f'Finding step log for {i_name} in branch: {current_branch}')
            else:
                # Its even, so we are in step, we start here!
                current_step = current_steps['.'.join(dot_path[:i + 1])]
                logger.debug(f'Finding step log for {i_name} in step: {current_step}')

        logger.debug(f'current branch : {current_branch}, current step {current_step}')
        if current_branch and current_step:
            return current_step, current_branch

        raise exceptions.StepLogNotFoundError(self.run_id, i_name)


# All outside modules should interact with dataclasses using the RunLogStore to promote extensibility
# If you want to customize dataclass, extend BaseRunLogStore and implement the methods as per the specification

class BaseRunLogStore:
    """
    The base class of a Run Log Store with many common methods implemented.

    Note: As a general guideline, do not extract anything from the config to set class level attributes.
          Integration patterns modify the config after init to change behaviors.
          Access config properties using getters/property of the class.
    """
    service_name = ''

    class Config(BaseModel):
        pass

    def __init__(self, config):
        config = config or {}
        self.config = self.Config(**config)

    def create_run_log(self, run_id: str, dag_hash: str = '', use_cached: bool = False,
                       tag: str = '', original_run_id: str = '', status: str = defaults.CREATED, **kwargs):
        """
        Creates a Run Log object by using the config

        Logically the method should do the following:
            * Creates a Run log
            * Adds it to the db
            * Return the log
        Raises:
            NotImplementedError: This is a base class and therefore has no default implementation
        """

        raise NotImplementedError

    def get_run_log_by_id(self, run_id: str, full: bool = True, **kwargs) -> RunLog:
        """
        Retrieves a Run log from the database using the config and the run_id

        Args:
            run_id (str): The run_id of the run
            full (bool): return the full run log store or only the RunLog object

        Returns:
            RunLog: The RunLog object identified by the run_id

        Logically the method should:
            * Returns the run_log defined by id from the data store defined by the config

        Raises:
            NotImplementedError: This is a base class and therefore has no default implementation
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """

        raise NotImplementedError

    def put_run_log(self, run_log: RunLog, **kwargs):
        """
        Puts the Run Log in the database as defined by the config

        Args:
            run_log (RunLog): The Run log of the run

        Logically the method should:
            Puts the run_log into the database

        Raises:
            NotImplementedError: This is a base class and therefore has no default implementation
        """
        raise NotImplementedError

    def get_parameters(self, run_id: str, **kwargs) -> dict:  # pylint: disable=unused-argument
        """
        Get the parameters from the Run log defined by the run_id

        Args:
            run_id (str): The run_id of the run

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Return the parameters as identified in the run_log

        Returns:
            dict: A dictionary of the run_log parameters
        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        run_log = self.get_run_log_by_id(run_id=run_id)
        return run_log.parameters

    def set_parameters(self, run_id: str, parameters: dict, **kwargs):  # pylint: disable=unused-argument
        """
        Update the parameters of the Run log with the new parameters

        This method would over-write the parameters, if the parameter exists in the run log already

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Update the parameters of the run_log
            * Call put_run_log(run_log) to put the run_log in the datastore

        Args:
            run_id (str): The run_id of the run
            parameters (dict): The parameters to update in the run log
        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
        """
        run_log = self.get_run_log_by_id(run_id=run_id)
        run_log.parameters.update(parameters)
        self.put_run_log(run_log=run_log)

    def get_run_config(self, run_id: str, **kwargs) -> dict:  # pylint: disable=unused-argument
        """
        Given a run_id, return the run_config used to perform the run.

        Args:
            run_id (str): The run_id of the run

        Returns:
            dict: The run config used for the run
        """

        run_log = self.get_run_log_by_id(run_id=run_id)
        return run_log.run_config

    def set_run_config(self, run_id: str, run_config: dict, **kwargs):  # pylint: disable=unused-argument
        """ Set the run config used to run the run_id

        Args:
            run_id (str): The run_id of the run
            run_config (dict): The run_config of the run
        """

        run_log = self.get_run_log_by_id(run_id=run_id)
        run_log.run_config.update(run_config)
        self.put_run_log(run_log=run_log)

    def create_step_log(self, name: str, internal_name: str, **kwargs):  # pylint: disable=unused-argument
        """
        Create a step log by the name and internal name

        The method does not update the Run Log with the step log at this point in time.
        This method is just an interface for external modules to create a step log


        Args:
            name (str): The friendly name of the step log
            internal_name (str): The internal naming of the step log. The internal naming is a dot path convention

        Returns:
            StepLog: A uncommitted step log object
        """
        logger.info(f'{self.service_name} Creating a Step Log: {name}')
        return StepLog(name=name, internal_name=internal_name, status=defaults.CREATED)

    def get_step_log(self, internal_name: str, run_id: str, **kwargs) -> StepLog:  # pylint: disable=unused-argument
        """
        Get a step log from the datastore for run_id and the internal naming of the step log

        The internal naming of the step log is a dot path convention.

        The method should:
            * Call get_run_log_by_id(run_id) to retrieve the run_log
            * Identify the step location by decoding the internal naming
            * Return the step log

        Args:
            internal_name (str): The internal name of the step log
            run_id (str): The run_id of the run

        Returns:
            StepLog: The step log object for the step defined by the internal naming and run_id

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
            StepLogNotFoundError: If the step log for internal_name is not found in the datastore for run_id
        """
        logger.info(f'{self.service_name} Getting the step log: {internal_name} of {run_id}')
        run_log = self.get_run_log_by_id(run_id=run_id)
        step_log, _ = run_log.search_step_by_internal_name(internal_name)
        return step_log

    def add_step_log(self, step_log: StepLog, run_id: str, **kwargs):  # pylint: disable=unused-argument
        """
        Add the step log in the run log as identified by the run_id in the datastore

        The method should:
             * Call get_run_log_by_id(run_id) to retrieve the run_log
             * Identify the branch to add the step by decoding the step_logs internal name
             * Add the step log to the identified branch log
             * Call put_run_log(run_log) to put the run_log in the datastore

        Args:
            step_log (StepLog): The Step log to add to the database
            run_id (str): The run id of the run

        Raises:
            RunLogNotFoundError: If the run log for run_id is not found in the datastore
            BranchLogNotFoundError: If the branch of the step log for internal_name is not found in the datastore
                                    for run_id
        """
        logger.info(f'{self.service_name} Adding the step log to DB: {step_log.name}')
        run_log = self.get_run_log_by_id(run_id=run_id)

        branch_to_add = '.'.join(step_log.internal_name.split('.')[:-1])
        branch, _ = run_log.search_branch_by_internal_name(branch_to_add)

        if branch is None:
            branch = run_log
        branch.steps[step_log.internal_name] = step_log
        self.put_run_log(run_log=run_log)

    def create_branch_log(self, internal_branch_name: str, **kwargs) -> BranchLog:  # pylint: disable=unused-argument
        """
        Creates a uncomitted branch log object by the internal name given

        Args:
            internal_branch_name (str): Creates a branch log by name internal_branch_name

        Returns:
            BranchLog: Uncommitted and initialized with defaults BranchLog object
        """
        # Create a new BranchLog
        logger.info(f'{self.service_name} Creating a Branch Log : {internal_branch_name}')
        return BranchLog(internal_name=internal_branch_name, status=defaults.CREATED)

    def get_branch_log(self, internal_branch_name: str, run_id: str, **kwargs) -> Union[BranchLog, RunLog]:  # pylint: disable=unused-argument
        """
        Returns the branch log by the internal branch name for the run id

        If the internal branch name is none, returns the run log

        Args:
            internal_branch_name (str): The internal branch name to retrieve.
            run_id (str): The run id of interest

        Returns:
            BranchLog: The branch log or the run log as requested.
        """
        run_log = self.get_run_log_by_id(run_id=run_id)
        if not internal_branch_name:
            return run_log
        branch, _ = run_log.search_branch_by_internal_name(internal_branch_name)
        return branch

    def add_branch_log(self, branch_log: Union[BranchLog, RunLog], run_id: str, **kwargs):  # pylint: disable=unused-argument
        """
        The method should:
        # Get the run log
        # Get the branch and step containing the branch
        # Add the branch to the step
        # Write the run_log

        The branch log could some times be a Run log and should be handled appropriately

        Args:
            branch_log (BranchLog): The branch log/run log to add to the database
            run_id (str): The run id to which the branch/run log is added
        """

        internal_branch_name = None

        if isinstance(branch_log, BranchLog):
            internal_branch_name = branch_log.internal_name

        if not internal_branch_name:
            self.put_run_log(branch_log)  # type: ignore # We are dealing with base dag here
            return

        run_log = self.get_run_log_by_id(run_id=run_id)

        step_name = '.'.join(internal_branch_name.split('.')[:-1])
        step, _ = run_log.search_step_by_internal_name(step_name)

        step.branches[internal_branch_name] = branch_log  # type: ignore
        self.put_run_log(run_log)

    def create_attempt_log(self, **kwargs) -> StepAttempt:  # pylint: disable=unused-argument
        """
        Returns an uncommitted step attempt log.

        Returns:
            StepAttempt: An uncommitted step attempt log
        """
        logger.info(f'{self.service_name} Creating an attempt log')
        return StepAttempt()

    def create_code_identity(self, **kwargs) -> CodeIdentity:  # pylint: disable=unused-argument
        """
        Creates an uncommitted Code identity class

        Returns:
            CodeIdentity: An uncommitted code identity class
        """
        logger.info(f'{self.service_name} Creating Code identity')
        return CodeIdentity()

    def create_data_catalog(self, name: str, **kwargs) -> DataCatalog:  # pylint: disable=unused-argument
        """
        Create a uncommitted data catalog object

        Args:
            name (str): The name of the data catalog item to put

        Returns:
            DataCatalog: The DataCatalog object.
        """
        logger.info(f'{self.service_name} Creating Data Catalog for {name}')
        return DataCatalog(name=name)


class BufferRunLogstore(BaseRunLogStore):
    """
    A in-memory run log store.

    This Run Log store will not persist any results.

    When to use:
     When testing some part of the pipeline.

    Do not use:
     When you need to compare between runs or in production set up

    This Run Log Store is concurrent write safe as it is in memory

    Example config:
    run_log:
      type: buffered

    """
    service_name = 'buffered'

    def __init__(self, config):
        super().__init__(config)
        self.run_log = None  # For a buffered Run Log, this is the database

    def create_run_log(self, run_id: str, dag_hash: str = '', use_cached: bool = False,
                       tag: str = '', original_run_id: str = '', status: str = defaults.CREATED, **kwargs) -> RunLog:
        # Creates a Run log
        # Adds it to the db
        # Return the log
        logger.info(f'{self.service_name} Creating a Run Log and adding it to DB')
        self.run_log = RunLog(run_id=run_id, dag_hash=dag_hash, use_cached=use_cached,
                              tag=tag, original_run_id=original_run_id, status=status)
        return self.run_log

    def get_run_log_by_id(self, run_id: str, full: bool = True, **kwargs):
        # Returns the run_log defined by id
        # Raises Exception if not found
        logger.info(f'{self.service_name} Getting the run log from DB for {run_id}')
        if self.run_log:
            return self.run_log

        raise exceptions.RunLogNotFoundError(run_id)

    def put_run_log(self, run_log: RunLog, **kwargs):
        # Puts the run_log into the database
        logger.info(f'{self.service_name} Putting the run log in the DB: {run_log.run_id}')
        self.run_log = run_log


class FileSystemRunLogstore(BaseRunLogStore):
    """
    In this type of Run Log store, we use a file system to store the JSON run log.

    Every single run is stored as a different file which makes it compatible across other store types.

    When to use:
        When locally testing a pipeline and have the need to compare across runs.
        Its fully featured and perfectly fine if your local environment is where you would do everyhing.

    Do not use:
        If you need parallelization on local, this run log would not support it.

    Example config:

    run_log:
      type: file-system
      config:
        log_folder: The folder to out the logs. Defaults to .run_log_store
    """
    service_name = 'file-system'

    class Config(BaseModel):
        log_folder: str = defaults.LOG_LOCATION_FOLDER

    @property
    def log_folder_name(self) -> str:
        return self.config.log_folder

    def write_to_folder(self, run_log: RunLog):
        """
        Write the run log to the folder

        Args:
            run_log (RunLog): The run log to be added to the database
        """
        write_to = self.log_folder_name
        utils.safe_make_dir(write_to)

        write_to_path = Path(write_to)
        run_id = run_log.run_id
        json_file_path = write_to_path / f'{run_id}.json'

        with json_file_path.open('w') as fw:
            json.dump(run_log.dict(), fw, ensure_ascii=True, indent=4)  # pylint: disable=no-member

    def get_from_folder(self, run_id: str) -> RunLog:
        """
        Look into the run log folder for the run log for the run id.

        If the run log does not exist, raise an exception. If it does, decode it
        as a RunLog and return it

        Args:
            run_id (str): The requested run id to retrieve the run log store

        Raises:
            FileNotFoundError: If the Run Log has not been found.

        Returns:
            RunLog: The decoded Run log
        """
        write_to = self.log_folder_name

        read_from_path = Path(write_to)
        json_file_path = read_from_path / f'{run_id}.json'

        if not json_file_path.exists():
            raise FileNotFoundError(f'Expected {json_file_path} is not present')

        with json_file_path.open('r') as fr:
            json_str = json.load(fr)
            run_log = RunLog(**json_str)  # pylint: disable=no-member
        return run_log

    def create_run_log(self, run_id: str, dag_hash: str = '', use_cached: bool = False,
                       tag: str = '', original_run_id: str = '', status: str = defaults.CREATED, **kwargs) -> RunLog:
        # Creates a Run log
        # Adds it to the db
        try:
            self.get_run_log_by_id(run_id=run_id, full=False)
            raise exceptions.RunLogExistsError(run_id=run_id)
        except exceptions.RunLogNotFoundError:
            pass

        logger.info(f'{self.service_name} Creating a Run Log for : {run_id}')
        run_log = RunLog(run_id=run_id, dag_hash=dag_hash, use_cached=use_cached,
                         tag=tag, original_run_id=original_run_id, status=status)
        self.write_to_folder(run_log)
        return run_log

    def get_run_log_by_id(self, run_id: str, full: bool = True, **kwargs) -> RunLog:
        # Returns the run_log defined by id
        # Raises Exception if not found
        try:
            logger.info(f'{self.service_name} Getting a Run Log for : {run_id}')
            run_log = self.get_from_folder(run_id)
            return run_log
        except FileNotFoundError as e:
            raise exceptions.RunLogNotFoundError(run_id) from e

    def put_run_log(self, run_log: RunLog, **kwargs):
        # Puts the run_log into the database
        logger.info(f'{self.service_name} Putting the run log in the DB: {run_log.run_id}')
        self.write_to_folder(run_log)



from enum import Enum

NAME = 'magnus'

# CLI settings
LOG_LEVEL = 'WARNING'


class EXECUTION_PLAN(Enum):
    # execution_plans
    pipeline = 1
    decorator = 2
    notebook = 3
    function = 4


# Config file environment variable
MAGNUS_CONFIG_FILE = 'MAGNUS_CONFIG_FILE'
MAGNUS_RUN_TAG = 'MAGNUS_RUN_TAG'

# Interaction settings
TRACK_PREFIX = 'MAGNUS_TRACK_'
PARAMETER_PREFIX = 'MAGNUS_PRM_'
VARIABLE_PREFIX = 'MAGNUS_VAR_'
ENV_RUN_ID = 'MAGNUS_RUN_ID'

# STATUS progression
# For Branch, CREATED -> PROCESSING -> SUCCESS OR FAIL
# For a step, CREATED -> TRIGGERED ->  PROCESSING -> SUCCESS OR FAIL
CREATED = 'CREATED'
PROCESSING = 'PROCESSING'
SUCCESS = 'SUCCESS'
FAIL = 'FAIL'
TRIGGERED = 'TRIGGERED'

# Node and Command settings
COMMAND_TYPE = 'python'
NODE_SPEC_FILE = 'node_spec.yaml'
COMMAND_FRIENDLY_CHARACTER = '%'

# Default services
DEFAULT_EXECUTOR = {
    'type': 'local'
}
DEFAULT_RUN_LOG_STORE = {
    'type': 'buffered'
}
DEFAULT_CATALOG = {
    'type': 'file-system'
}
DEFAULT_SECRETS = {
    'type': 'do-nothing'
}
DEFAULT_EXPERIMENT_TRACKER = {
    'type': 'do-nothing'
}

# Map state
MAP_PLACEHOLDER = 'map_variable_placeholder'

# Dag node
DAG_BRANCH_NAME = 'dag'

# RUN settings
RANDOM_RUN_ID_LEN = 6
MAX_TIME = 86400  # 1 day in seconds

# User extensions
USER_CONFIG_FILE = 'magnus-config.yaml'

# Executor settings
ENABLE_PARALLEL = False

# RUN log store settings
LOG_LOCATION_FOLDER = '.run_log_store'

# Dag node
DAG_BRANCH_NAME = 'dag'

# Data catalog settings
CATALOG_LOCATION_FOLDER = '.catalog'
COMPUTE_DATA_FOLDER = 'data'

# Secrets settings
DOTENV_FILE_LOCATION = '.env'

# AWS settings
AWS_REGION = 'eu-west-1'

# Docker settings
DOCKERFILE_NAME = 'Dockerfile'
DOCKERFILE_CONTENT = r"""# Python 3.6 Image without Dependecies
FROM python:3.8

LABEL maintainer="vijay.vammi@astrazeneca.com"

RUN apt-get update && apt-get install -y --no-install-recommends \
    git \
    && rm -rf /var/lib/apt/lists/*

${INSTALL_STYLE}

ENV VIRTUAL_ENV=/opt/venv
RUN python -m virtualenv --python=/usr/local/bin/python $VIRTUAL_ENV
ENV PATH="$VIRTUAL_ENV/bin:$PATH"

${COPY_CONTENT}
WORKDIR /app

${INSTALL_REQUIREMENTS}
"""
GIT_ARCHIVE_NAME = 'git_tracked'
LEN_SHA_FOR_TAG = 8



import logging
import os
import shutil
from pathlib import Path
from typing import List

from pydantic import BaseModel

from magnus import defaults, utils

logger = logging.getLogger(defaults.NAME)


def get_run_log_store():
    """
    This method should be called after the executor module has been populated with all the systems.

    This method retrieves the run log store from the global executor.

    Returns:
        object: The run log store
    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel
    return global_executor.run_log_store


def is_catalog_out_of_sync(catalog, synced_catalogs=None) -> bool:
    """
    Check if the catalog items are out of sync from already cataloged objects.
    If they are, return False.
    If the object does not exist or synced catalog does not exist, return True
    """
    if not synced_catalogs:
        return True  # If nothing has been synced in the past

    for synced_catalog in synced_catalogs:
        if synced_catalog.catalog_relative_path == catalog.catalog_relative_path:
            if synced_catalog.data_hash == catalog.data_hash:
                return False
            return True

    return True  # The object does not exist, sync it


class BaseCatalog:
    """
    Base Catalog class definition.

    All implementations of the catalog handler should inherit and extend this class.

    Note: As a general guideline, do not extract anything from the config to set class level attributes.
          Integration patterns modify the config after init to change behaviors.
          Access config properties using getters/property of the class.
    """
    service_name = ''

    class Config(BaseModel):
        compute_data_folder: str = defaults.COMPUTE_DATA_FOLDER

    def __init__(self, config: dict, **kwargs):  # pylint: disable=unused-argument
        config = config or {}
        self.config = self.Config(**config)

    @property
    def compute_data_folder(self) -> str:
        """
        Returns the compute data folder defined as per the config of the catalog.

        Returns:
            [str]: The compute data folder as defined or defaults to magnus default 'data/'
        """
        return self.config.compute_data_folder

    def get(self, name: str, run_id: str, compute_data_folder=None, **kwargs) -> List[object]:
        # pylint: disable=unused-argument
        """
        Get the catalog item by 'name' for the 'run id' and store it in compute data folder.

        The catalog location should have been created before you can get from it.

        Args:
            name (str): The name of the catalog item
            run_id (str): The run_id of the run.
            compute_data_folder (str, optional): The compute data folder. Defaults to magnus default (data/)

        Raises:
            NotImplementedError: Base class, hence not implemented

        Returns:
            List(object) : A list of catalog objects
        """
        raise NotImplementedError

    def put(self, name: str, run_id: str, compute_data_folder=None, synced_catalogs=None, **kwargs) -> List[object]:
        # pylint: disable=unused-argument
        """
        Put the file by 'name' from the 'compute_data_folder' in the catalog for the run_id.

        If previous syncing has happened and the file has not been changed, we do not sync again.

        Args:
            name (str): The name of the catalog item.
            run_id (str): The run_id of the run.
            compute_data_folder (str, optional): The compute data folder. Defaults to magnus default (data/)
            synced_catalogs (dict, optional): Any previously synced catalogs. Defaults to None.

        Raises:
            NotImplementedError: Base class, hence not implemented

        Returns:
            List(object) : A list of catalog objects
        """
        raise NotImplementedError

    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Given run_id of a previous run, sync them to the catalog of the run given by run_id

        Args:
            previous_run_id (str): The run id of the previous run
            run_id (str): The run_id to which the data catalogs should be synced to.

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        raise NotImplementedError


class DoNothingCatalog(BaseCatalog):
    """
    A Catalog handler that does nothing.

    Example config:

    catalog:
      type: do-nothing

    """
    service_name = 'do-nothing'

    def get(self, name: str, run_id: str, compute_data_folder=None, **kwargs) -> List[object]:
        """
        Does nothing
        """
        logger.info('Using a do-nothing catalog, doing nothing in get')
        return []

    def put(self, name: str, run_id: str, compute_data_folder=None, synced_catalogs=None, **kwargs) -> List[object]:
        """
        Does nothing
        """
        logger.info('Using a do-nothing catalog, doing nothing in put')
        return []

    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Does nothing
        """
        logger.info('Using a do-nothing catalog, doing nothing while sync between runs')
        ...


class FileSystemCatalog(BaseCatalog):
    """
    A Catalog handler that uses the local file system for cataloging.

    Note: Do not use this if the steps of the pipeline run on different compute environments.

    Example config:

    catalog:
      type: file-system
      config:
        catalog_location: The location to store the catalog.
        compute_data_folder: The folder to source the data from.

    """
    # TODO think of compression algorithms to save disk space, ZipFile
    service_name = 'file-system'

    class Config(BaseCatalog.Config):
        catalog_location: str = defaults.CATALOG_LOCATION_FOLDER

    @property
    def catalog_location(self) -> str:
        """
        Get the catalog location from the config.
        If its not defined, use the magnus default

        Returns:
            str: The catalog location as defined by the config or magnus default '.catalog'
        """
        return self.config.catalog_location  # type: ignore

    def get(self, name: str, run_id: str, compute_data_folder=None, **kwargs) -> List[object]:
        """
        Get the file by matching glob pattern to the name

        Args:
            name ([str]): A glob matching the file name
            run_id ([str]): The run id

        Raises:
            Exception: If the catalog location does not exist

        Returns:
            List(object) : A list of catalog objects
        """
        logger.info(f'Using the {self.service_name} catalog and trying to get {name} for run_id: {run_id}')

        copy_to = self.compute_data_folder
        if compute_data_folder:
            copy_to = compute_data_folder

        copy_to = Path(copy_to)  # type: ignore

        if not utils.does_dir_exist(copy_to):
            msg = (
                f'Expected compute data folder to be present at: {copy_to} but not found. \n'
                'Note: Magnus does not create the compute data folder for you. Please ensure that the folder exists.\n'
            )
            raise Exception(msg)

        catalog_location = self.catalog_location
        run_catalog = Path(catalog_location) / run_id / copy_to

        logger.debug(f'Copying objects to {copy_to} from the run catalog location of {run_catalog}')

        if not utils.does_dir_exist(run_catalog):
            msg = (
                f'Expected Catalog to be present at: {run_catalog} but not found.\n'
                'Note: Please make sure that some data was put in the catalog before trying to get from it.\n'
            )
            raise Exception(msg)

        # Iterate through the contents of the run_catalog and copy the files that fit the name pattern
        # We should also return a list of data hashes
        glob_files = run_catalog.glob(name)
        logger.debug(f'Glob identified {glob_files} as matches to from the catalog location: {run_catalog}')

        data_catalogs = []
        run_log_store = get_run_log_store()
        for file in glob_files:
            if file.is_dir():
                # Need not add a data catalog for the folder
                continue

            relative_file_path = file.relative_to(run_catalog)

            data_catalog = run_log_store.create_data_catalog(str(relative_file_path))
            data_catalog.catalog_handler_location = catalog_location
            data_catalog.catalog_relative_path = str(relative_file_path)
            data_catalog.data_hash = utils.get_data_hash(str(file))
            data_catalog.stage = 'get'
            data_catalogs.append(data_catalog)

            # Make the directory in the data folder if required
            Path(copy_to / relative_file_path.parent).mkdir(parents=True, exist_ok=True)
            shutil.copy(file, copy_to / relative_file_path)

            logger.info(f'Copied {file} from {run_catalog} to {copy_to}')

        return data_catalogs

    def put(self, name: str, run_id: str, compute_data_folder=None, synced_catalogs=None, **kwargs) -> List[object]:
        """
        Put the files matching the glob pattern into the catalog.

        If previously synced catalogs are provided, and no changes were observed, we do not sync them.

        Args:
            name (str): The glob pattern of the files to catalog
            run_id (str): The run id of the run
            compute_data_folder (str, optional): The compute data folder to sync from. Defaults to settings default.
            synced_catalogs (dict, optional): dictionary of previously synced catalogs. Defaults to None.

        Raises:
            Exception: If the compute data folder does not exist.

        Returns:
            List(object) : A list of catalog objects
        """
        logger.info(f'Using the {self.service_name} catalog and trying to put {name} for run_id: {run_id}')

        copy_from = self.compute_data_folder
        if compute_data_folder:
            copy_from = compute_data_folder
        copy_from = Path(copy_from)  # type: ignore

        catalog_location = self.catalog_location
        run_catalog = Path(catalog_location) / run_id
        utils.safe_make_dir(run_catalog)

        logger.debug(f'Copying objects from {copy_from} to the run catalog location of {run_catalog}')

        if not utils.does_dir_exist(copy_from):
            msg = (
                f'Expected compute data folder to be present at: {compute_data_folder} but not found. \n'
                'Note: Magnus does not create the compute data folder for you. Please ensure that the folder exists.\n'
            )
            raise Exception(msg)

        # Iterate through the contents of copy_from and if the name matches, we move them to the run_catalog
        # We should also return a list of datastore.DataCatalog items

        glob_files = copy_from.glob(name)  # type: ignore
        logger.debug(f'Glob identified {glob_files} as matches to from the compute data folder: {copy_from}')

        data_catalogs = []
        run_log_store = get_run_log_store()
        for file in glob_files:
            if file.is_dir():
                # Need not add a data catalog for the folder
                continue

            relative_file_path = file.relative_to('.')

            data_catalog = run_log_store.create_data_catalog(str(relative_file_path))
            data_catalog.catalog_handler_location = catalog_location
            data_catalog.catalog_relative_path = run_id + os.sep + str(relative_file_path)
            data_catalog.data_hash = utils.get_data_hash(str(file))
            data_catalog.stage = 'put'
            data_catalogs.append(data_catalog)

            if is_catalog_out_of_sync(data_catalog, synced_catalogs):
                logger.info(f'{data_catalog.name} was found to be changed, syncing')

                # Make the directory in the catalog if required
                Path(run_catalog / relative_file_path.parent).mkdir(parents=True, exist_ok=True)
                shutil.copy(file, run_catalog / relative_file_path)
            else:
                logger.info(f'{data_catalog.name} was found to be unchanged, ignoring syncing')
        return data_catalogs

    def sync_between_runs(self, previous_run_id: str, run_id: str):
        """
        Given the previous run id, sync the catalogs between the current one and previous

        Args:
            previous_run_id (str): The previous run id to sync the catalogs from
            run_id (str): The run_id to which the data catalogs should be synced to.

        Raises:
            Exception: If the previous run log does not exist in the catalog

        """
        logger.info(f'Using the {self.service_name} catalog and syncing catalogs'
                    'between old: {previous_run_id} to new: {run_id}')

        catalog_location = Path(self.catalog_location)
        run_catalog = catalog_location / run_id
        utils.safe_make_dir(run_catalog)

        if not utils.does_dir_exist(catalog_location / previous_run_id):
            msg = (
                f'Catalogs from previous run : {previous_run_id} are not found.\n'
                'Note: Please provision the catalog objects generated by previous run in the same catalog location'
                ' as the current run, even if the catalog handler for the previous run was different')
            raise Exception(msg)

        cataloged_files = (catalog_location / previous_run_id).glob('**/**')

        for cataloged_file in cataloged_files:
            shutil.copy(cataloged_file, run_catalog)
            logger.info(f'Copied file from: {cataloged_file} to {run_catalog}')



class RunLogExistsError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id):
        super().__init__()
        self.message = f'Run id for {run_id} is already found in the datastore'


class RunLogNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id):
        super().__init__()
        self.message = f'Run id for {run_id} is not found in the datastore'


class StepLogNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id, name):
        super().__init__()
        self.message = f'Step log for {name} is not found in the datastore for Run id: {run_id}'


class BranchLogNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, run_id, name):
        super().__init__()
        self.message = f'Branch log for {name} is not found in the datastore for Run id: {run_id}'


class NodeNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, name):
        super().__init__()
        self.message = f'Node of name {name} is not found the graph'


class BranchNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, name):
        super().__init__()
        self.message = f'Branch of name {name} is not found the graph'


class UnSupportedModeError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, name, operation):
        super().__init__()
        self.message = f'Mode of type {name} is not supporterted for operation {operation}'


class SecretNotFoundError(Exception):  # pragma: no cover
    """
    Exception class
    Args:
        Exception ([type]): [description]
    """

    def __init__(self, secret_name, secret_setting):
        super().__init__()
        self.message = f'No secret found by name:{secret_name} in {secret_setting}'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magnus']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'click-plugins>=1.1.1,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'ruamel.yaml',
 'ruamel.yaml.clib',
 'stevedore>=3.5.0,<4.0.0',
 'yachalk']

extras_require = \
{'docker': ['docker'], 'notebook': ['ploomber-engine>=0.0.19,<0.0.20']}

entry_points = \
{'console_scripts': ['magnus = magnus.cli:cli'],
 'magnus.catalog.BaseCatalog': ['do-nothing = magnus.catalog:DoNothingCatalog',
                                'file-system = '
                                'magnus.catalog:FileSystemCatalog'],
 'magnus.datastore.BaseRunLogStore': ['buffered = '
                                      'magnus.datastore:BufferRunLogstore',
                                      'file-system = '
                                      'magnus.datastore:FileSystemRunLogstore'],
 'magnus.executor.BaseExecutor': ['demo-renderer = '
                                  'magnus.executor:DemoRenderer',
                                  'local = magnus.executor:LocalExecutor',
                                  'local-container = '
                                  'magnus.executor:LocalContainerExecutor'],
 'magnus.experiment_tracker.BaseExperimentTracker': ['do-nothing = '
                                                     'magnus.experiment_tracker:DoNothingTracker',
                                                     'mlflow = '
                                                     'magnus.experiment_tracker:MLFlowExperimentTracker'],
 'magnus.integration.BaseIntegration': ['demo-renderer-run_log_store-buffered '
                                        '= '
                                        'magnus.integration:DemoRenderBufferedRunLogStore',
                                        'local-catalog-do-nothing = '
                                        'magnus.integration:LocalDoNothingCatalog',
                                        'local-container-catalog-do-nothing = '
                                        'magnus.integration:LocalContainerDoNothingCatalog',
                                        'local-container-catalog-file-system = '
                                        'magnus.integration:LocalContainerComputeFileSystemCatalog',
                                        'local-container-run_log_store-buffered '
                                        '= '
                                        'magnus.integration:LocalContainerComputeBufferedRunLogStore',
                                        'local-container-run_log_store-file-system '
                                        '= '
                                        'magnus.integration:LocalContainerComputeFileSystemRunLogstore',
                                        'local-container-secrets-dotenv = '
                                        'magnus.integration:LocalContainerComputeDotEnvSecrets',
                                        'local-container-secrets-environment = '
                                        'magnus.integration:LocalContainerComputeEnvSecretsManager',
                                        'local-run_log_store-buffered = '
                                        'magnus.integration:LocalComputeBufferedRunLogStore',
                                        'local-run_log_store-file-system = '
                                        'magnus.integration:LocalComputeFileSystemRunLogStore'],
 'magnus.nodes.BaseNode': ['as-is = magnus.nodes:AsISNode',
                           'dag = magnus.nodes:DagNode',
                           'fail = magnus.nodes:FailNode',
                           'map = magnus.nodes:MapNode',
                           'parallel = magnus.nodes:ParallelNode',
                           'success = magnus.nodes:SuccessNode',
                           'task = magnus.nodes:TaskNode'],
 'magnus.secrets.BaseSecrets': ['do-nothing = '
                                'magnus.secrets:DoNothingSecretManager',
                                'dotenv = magnus.secrets:DotEnvSecrets',
                                'env-secrets-manager = '
                                'magnus.secrets:EnvSecretsManager'],
 'magnus.tasks.BaseTaskType': ['notebook = magnus.tasks:NotebookTaskType',
                               'python = magnus.tasks:PythonTaskType',
                               'python-function = '
                               'magnus.tasks:PythonFunctionType',
                               'python-lambda = '
                               'magnus.tasks:PythonLambdaTaskType',
                               'shell = magnus.tasks:ShellTaskType']}

setup_kwargs = {
    'name': 'magnus',
    'version': '0.4.0',
    'description': 'A Compute agnostic pipelining software',
    'long_description': '# Hello from magnus\n\n\n![logo](docs/assets/logo1.png)\n---\n\n**Magnus** is a *thin* layer of abstraction over the underlying infrastructure to enable data scientist and\nmachine learning engineers. It provides:\n\n- A way to execute Jupyter notebooks/python functions in local or remote platforms.\n- A framework to define complex pipelines via YAML or Python SDK.\n- Robust and *automatic* logging to ensure maximum reproducibility of experiments.\n- A framework to interact with secret managers ranging from environment variables to other vendors.\n- Interactions with various experiment tracking tools.\n\n## What does **thin** mean?\n\n- We really have no say in what happens within your notebooks or python functions.\n- We do not dictate how the infrastructure should be configured as long as it satisfies some *basic* criteria.\n    - The underlying infrastructure should support container execution and an orchestration framework.\n    - Some way to handle secrets either via environment variables or secrets manager.\n    - A blob storage or some way to store your intermediate artifacts.\n    - A database or blob storage to store logs.\n- We have no opinion of how your structure your project.\n- We do not creep into your CI/CD practices but it is your responsibility to provide the same environment where ever\nthe execution happens. This is usually via git, virtual environment manager and docker.\n- We transpile to the orchestration framework that is used by your teams to do the heavy lifting.\n\n## What does it do?\n\n\n![works](docs/assets/work.png)\n\n### Shift Left\n\nMagnus provides patterns typically used in production environments even in the development phase.\n\n- Reduces the need for code refactoring during production phase of the project.\n- Enables best practices and understanding of infrastructure patterns.\n- Run the same code on your local machines or in production environments.\n\n:sparkles::sparkles:Happy Experimenting!!:sparkles::sparkles:\n\n\n## Documentation\n\n[More details about the project and how to use it available here](https://astrazeneca.github.io/magnus-core/).\n\n## Installation\n\n\nThe minimum python version that magnus supports is 3.8\n## pip\n\nmagnus is a python package and should be installed as any other.\n\n```shell\npip install magnus\n```\n\nWe recommend that you install magnus in a virtual environment specific to the project and also poetry for your\napplication development.\n\nThe command to install in a poetry managed virtual environment\n\n```\npoetry add magnus\n```\n\n## Example Run\n\nTo give you a flavour of how magnus works, lets create a simple pipeline.\n\nCopy the contents of this yaml into getting-started.yaml or alternatively in a python file if you are using the SDK.\n\n---\n!!! Note\n\n   The below execution would create a folder called \'data\' in the current working directory.\n   The command as given should work in linux/macOS but for windows, please change accordingly.\n\n---\n\n``` yaml\ndag:\n  description: Getting started\n  start_at: step parameters\n  steps:\n    step parameters:\n      type: task\n      command_type: python-lambda\n      command: "lambda x: {\'x\': int(x) + 1}"\n      next: step shell\n    step shell:\n      type: task\n      command_type: shell\n      command: mkdir data ; env >> data/data.txt # For Linux/macOS\n      next: success\n      catalog:\n        put:\n          - "*"\n    success:\n      type: success\n    fail:\n      type: fail\n```\n\nThe same could also be defined via a Python SDK.\n\n```python\n\n#in pipeline.py\nfrom magnus import Pipeline, Task\n\ndef pipeline():\n    first = Task(name=\'step parameters\', command="lambda x: {\'x\': int(x) + 1}", command_type=\'python-lambda\')\n    second = Task(name=\'step shell\', command=\'mkdir data ; env >> data/data.txt\',\n                  command_type=\'shell\', catalog={\'put\': \'*\'})\n\n    pipeline = Pipeline(name=\'getting_started\')\n    pipeline.construct([first, second])\n    pipeline.execute(parameters_file=\'parameters.yaml\')\n\nif __name__ == \'__main__\':\n    pipeline()\n\n```\n\nSince the pipeline expects a parameter ```x```, lets provide that using ```parameters.yaml```\n\n```yaml\nx: 3\n```\n\n\nAnd let\'s run the pipeline using:\n``` shell\n magnus execute --file getting-started.yaml --parameters-file parameters.yaml\n```\n\nIf you are using the python SDK:\n\n```\npoetry run python pipeline.py\n```\n\nYou should see a list of warnings but your terminal output should look something similar to this:\n\n``` json\n{\n    "run_id": "20230131195647",\n    "dag_hash": "",\n    "use_cached": false,\n    "tag": "",\n    "original_run_id": "",\n    "status": "SUCCESS",\n    "steps": {\n        "step parameters": {\n            "name": "step parameters",\n            "internal_name": "step parameters",\n            "status": "SUCCESS",\n            "step_type": "task",\n            "message": "",\n            "mock": false,\n            "code_identities": [\n                {\n                    "code_identifier": "e15d1374aac217f649972d11fe772e61b5a2478d",\n                    "code_identifier_type": "git",\n                    "code_identifier_dependable": true,\n                    "code_identifier_url": "INTENTIONALLY REMOVED",\n                    "code_identifier_message": ""\n                }\n            ],\n            "attempts": [\n                {\n                    "attempt_number": 0,\n                    "start_time": "2023-01-31 19:56:55.007931",\n                    "end_time": "2023-01-31 19:56:55.009273",\n                    "duration": "0:00:00.001342",\n                    "status": "SUCCESS",\n                    "message": ""\n                }\n            ],\n            "user_defined_metrics": {},\n            "branches": {},\n            "data_catalog": []\n        },\n        "step shell": {\n            "name": "step shell",\n            "internal_name": "step shell",\n            "status": "SUCCESS",\n            "step_type": "task",\n            "message": "",\n            "mock": false,\n            "code_identities": [\n                {\n                    "code_identifier": "e15d1374aac217f649972d11fe772e61b5a2478d",\n                    "code_identifier_type": "git",\n                    "code_identifier_dependable": true,\n                    "code_identifier_url": "INTENTIONALLY REMOVED",\n                    "code_identifier_message": ""\n                }\n            ],\n            "attempts": [\n                {\n                    "attempt_number": 0,\n                    "start_time": "2023-01-31 19:56:55.128697",\n                    "end_time": "2023-01-31 19:56:55.150878",\n                    "duration": "0:00:00.022181",\n                    "status": "SUCCESS",\n                    "message": ""\n                }\n            ],\n            "user_defined_metrics": {},\n            "branches": {},\n            "data_catalog": [\n                {\n                    "name": "data/data.txt",\n                    "data_hash": "7e91b0a9ff8841a3b5bf2c711f58bcc0cbb6a7f85b9bc92aa65e78cdda59a96e",\n                    "catalog_relative_path": "20230131195647/data/data.txt",\n                    "catalog_handler_location": ".catalog",\n                    "stage": "put"\n                }\n            ]\n        },\n        "success": {\n            "name": "success",\n            "internal_name": "success",\n            "status": "SUCCESS",\n            "step_type": "success",\n            "message": "",\n            "mock": false,\n            "code_identities": [\n                {\n                    "code_identifier": "e15d1374aac217f649972d11fe772e61b5a2478d",\n                    "code_identifier_type": "git",\n                    "code_identifier_dependable": true,\n                    "code_identifier_url": "INTENTIONALLY REMOVED",\n                    "code_identifier_message": ""\n                }\n            ],\n            "attempts": [\n                {\n                    "attempt_number": 0,\n                    "start_time": "2023-01-31 19:56:55.239877",\n                    "end_time": "2023-01-31 19:56:55.240116",\n                    "duration": "0:00:00.000239",\n                    "status": "SUCCESS",\n                    "message": ""\n                }\n            ],\n            "user_defined_metrics": {},\n            "branches": {},\n            "data_catalog": []\n        }\n    },\n    "parameters": {\n        "x": 4\n    },\n    "run_config": {\n        "executor": {\n            "type": "local",\n            "config": {\n                "enable_parallel": false,\n                "placeholders": {}\n            }\n        },\n        "run_log_store": {\n            "type": "buffered",\n            "config": {}\n        },\n        "catalog": {\n            "type": "file-system",\n            "config": {\n                "compute_data_folder": "data",\n                "catalog_location": ".catalog"\n            }\n        },\n        "secrets": {\n            "type": "do-nothing",\n            "config": {}\n        },\n        "experiment_tracker": {\n            "type": "do-nothing",\n            "config": {}\n        },\n        "variables": {},\n        "pipeline": {\n            "start_at": "step parameters",\n            "name": "getting_started",\n            "description": "",\n            "max_time": 86400,\n            "steps": {\n                "step parameters": {\n                    "mode_config": {},\n                    "next_node": "step shell",\n                    "command": "lambda x: {\'x\': int(x) + 1}",\n                    "command_type": "python-lambda",\n                    "command_config": {},\n                    "catalog": {},\n                    "retry": 1,\n                    "on_failure": "",\n                    "type": "task"\n                },\n                "step shell": {\n                    "mode_config": {},\n                    "next_node": "success",\n                    "command": "mkdir data ; env >> data/data.txt",\n                    "command_type": "shell",\n                    "command_config": {},\n                    "catalog": {\n                        "put": "*"\n                    },\n                    "retry": 1,\n                    "on_failure": "",\n                    "type": "task"\n                },\n                "success": {\n                    "mode_config": {},\n                    "type": "success"\n                },\n                "fail": {\n                    "mode_config": {},\n                    "type": "fail"\n                }\n            }\n        }\n    }\n}\n```\n\nYou should see that ```data``` folder being created with a file called ```data.txt``` in it.\nThis is according to the command in ```step shell```.\n\nYou should also see a folder ```.catalog``` being created with a single folder corresponding to the run_id of this run.\n\nTo understand more about the input and output, please head over to the\n[documentation](https://project-magnus.github.io/magnus-core/).\n',
    'author': 'Vijay Vammi',
    'author_email': 'vijay.vammi@astrazeneca.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AstraZeneca/magnus-core',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import functools
import json
import logging
import os
from logging.config import fileConfig
from pathlib import Path
from types import FunctionType
from typing import Dict, List, Optional, Union

from pkg_resources import resource_filename

from magnus import defaults, exceptions, graph, nodes, pipeline, utils

logger = logging.getLogger(defaults.NAME)


def track_this(step: int = 0, **kwargs):
    """
    Set up the keyword args as environment variables for tracking purposes as
    part of the run.

    For every key-value pair found in kwargs, we set up an environmental variable of
    MAGNUS_TRACK_key_{step} = json.dumps(value)

    If step=0, we ignore step for magnus purposes.

    Args:
        kwargs (dict): The dictionary of key value pairs to track.
    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel

    prefix = defaults.TRACK_PREFIX

    if step:
        prefix += f'{str(step)}_'

    for key, value in kwargs.items():
        logger.info(f'Tracking {key} with value: {value}')
        os.environ[prefix + key] = json.dumps(value)
        global_executor.experiment_tracker.set_metric(key, value, step=step)  # type: ignore


def store_parameter(update: bool = True, **kwargs: dict):
    """
    Set up the keyword args as environment variables for parameters tracking
    purposes as part pf the run.

    If update_existing is True, we override the current value if the parameter already exists.

    For every key-value pair found in kwargs, we set up an environmental variable of
    MAGNUS_PRM_key = json.dumps(value)
    """
    for key, value in kwargs.items():
        logger.info(f'Storing parameter {key} with value: {value}')
        environ_key = defaults.PARAMETER_PREFIX + key

        if environ_key in os.environ and not update:
            return

        os.environ[environ_key] = json.dumps(value)


def get_parameter(key=None) -> Union[str, dict]:
    """
    Get the parameter set as part of the user convenience function.

    We do not remove the parameter from the environment in this phase as
    as the function execution has not been completed.

    Returns all the parameters, if no key was provided.

    Args:
        key (str, optional): The parameter key to retrieve. Defaults to None.

    Raises:
        Exception: If the mentioned key was not part of the paramters

    Returns:
        Union[str, dict]: A single value of str or a dictionary if no key was specified
    """
    parameters = utils.get_user_set_parameters(remove=False)
    if not key:
        return parameters
    if key not in parameters:
        raise Exception(f'Parameter {key} is not set before')
    return parameters[key]


def get_secret(secret_name: str = None) -> str:
    """
    Get a secret by the name from the secrets manager

    Args:
        secret_name (str): The name of the secret to get. Defaults to None.

    Returns:
        str: The secret from the secrets manager, if exists. If the requested secret was None, we return all.
        Otherwise, raises exception.

    Raises:
        exceptions.SecretNotFoundError: Secret not found in the secrets manager.
    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel
    secrets_handler = global_executor.secrets_handler  # type: ignore

    try:
        return secrets_handler.get(name=secret_name)
    except exceptions.SecretNotFoundError:
        logger.exception(f'No secret by the name {secret_name} found in the store')
        raise


def get_from_catalog(name: str, destination_folder: str = None):
    """
    A convenience interaction function to get file from the catalog and place it in the destination folder

    Note: We do not perform any kind of serialization/deserialization in this way.
    Args:
        name (str): The name of the file to get from the catalog
        destination_folder (None): The place to put the file. defaults to compute data folder

    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel

    if not destination_folder:
        destination_folder = global_executor.catalog_handler.compute_data_folder  # type: ignore

    data_catalog = global_executor.catalog_handler.get(name, run_id=global_executor.run_id,  # type: ignore
                                                       compute_data_folder=destination_folder)

    if global_executor.context_step_log:  # type: ignore
        global_executor.context_step_log.add_data_catalogs(data_catalog)  # type: ignore
    else:
        logger.warning("Step log context was not found during interaction! The step log will miss the record")


def put_in_catalog(filepath: str):
    """
    A convenience interaction function to put the file in the catalog.

    Note: We do not perform any kind of serialization/deserialization in this way.

    Args:
        filepath (str): The path of the file to put in the catalog
    """
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel

    file_path = Path(filepath)

    data_catalog = global_executor.catalog_handler.put(file_path.name, run_id=global_executor.run_id,  # type: ignore
                                                       compute_data_folder=file_path.parent)
    if global_executor.context_step_log:  # type: ignore
        global_executor.context_step_log.add_data_catalogs(data_catalog)  # type: ignore
    else:
        logger.warning("Step log context was not found during interaction! The step log will miss the record")


def get_run_id() -> str:
    """
    Returns the run_id of the current run
    """
    return os.environ.get(defaults.ENV_RUN_ID, '')


def get_tag() -> str:
    """
    Returns the tag from the environment.

    Returns:
        str: The tag if provided for the run, otherwise None
    """
    return os.environ.get(defaults.MAGNUS_RUN_TAG, '')


def get_experiment_tracker_context():
    from magnus.pipeline import \
        global_executor  # pylint: disable=import-outside-toplevel

    experiment_tracker = global_executor.experiment_tracker
    return experiment_tracker.client_context


class step(object):

    def __init__(
            self, name: Union[str, FunctionType],
            catalog_config: dict = None, magnus_config: str = None,
            parameters_file: str = None):
        """
        This decorator could be used to make the function within the scope of magnus.

        Since we are not orchestrating, it is expected that resource management happens outside this scope.

        Args:
            name (str, callable): The name of the step. The step log would have the same name
            catalog_config (dict): The configuration of the catalog per step.
            magnus_config (str): The name of the file having the magnus config, defaults to None.
        """
        if isinstance(name, FunctionType):
            name = name()

        self.name = name
        self.catalog_config = catalog_config
        self.active = True  # Check if we are executing the function via pipeline

        if pipeline.global_executor \
                and pipeline.global_executor.execution_plan != defaults.EXECUTION_PLAN.decorator:
            self.active = False
            return

        self.executor = pipeline.prepare_configurations(
            configuration_file=magnus_config, parameters_file=parameters_file)

        self.executor.execution_plan = defaults.EXECUTION_PLAN.decorator
        run_id = self.executor.step_decorator_run_id
        if not run_id:
            msg = (
                f'Step decorator expects run id from environment.'
            )
            raise Exception(msg)

        self.executor.run_id = run_id
        utils.set_magnus_environment_variables(run_id=run_id, configuration_file=magnus_config, tag=get_tag())

        try:
            # Try to get it if previous steps have created it
            # TODO: Can call the set_up_runlog now.
            run_log = self.executor.run_log_store.get_run_log_by_id(self.executor.run_id)
            if run_log.status in [defaults.FAIL, defaults.SUCCESS]:  # TODO: Remove this in preference to defaults
                """
                This check is mostly useless as we do not know when the graph ends as they are created dynamically.
                This only prevents from using a run_id which has reached a final state.
                #TODO: There is a need to create a status called step_success
                """
                msg = (
                    f'The run_log for run_id: {run_id} already exists and is in {run_log.status} state.'
                    ' Make sure that this was not run before.'
                )
                raise Exception(msg)
        except exceptions.RunLogNotFoundError:
            # Create one if they are not created
            self.executor._set_up_run_log()

    def __call__(self, func):
        """
        The function is converted into a node and called via the magnus framework.
        """
        @functools.wraps(func)
        def wrapped_f(*args, **kwargs):
            if not self.active:
                # If we are not running via decorator, execute the function
                return func(*args, **kwargs)

            step_config = {
                'command': func,
                'command_type': 'python-function',
                'type': 'task',
                'next': 'not defined',
                'catalog': self.catalog_config
            }
            node = graph.create_node(name=self.name, step_config=step_config)
            self.executor.execute_from_graph(node=node)
            run_log = self.executor.run_log_store.get_run_log_by_id(run_id=self.executor.run_id, full=False)
            # TODO: If the previous step succeeded, make the status of the run log step_success
            print(json.dumps(run_log.dict(), indent=4))
        return wrapped_f


class Task:
    def __init__(self, name: str, command: Union[str, FunctionType], command_type: str = defaults.COMMAND_TYPE,
                 command_config: Optional[dict] = None, catalog: Optional[dict] = None,
                 mode_config: Optional[dict] = None, retry: int = 1, on_failure: str = '',
                 next_node: str = ''):
        self.name = name
        self.command = command
        self.command_type = command_type
        self.command_config = command_config or {}
        self.catalog = catalog or {}
        self.mode_config = mode_config or {}
        self.retry = retry
        self.on_failure = on_failure
        self.next_node = next_node or "success"
        self.node: Optional[nodes.BaseNode] = None

    def _construct_node(self):
        # TODO: The below has issues if the function and the pipeline are in the same module
        # Something to do with __main__ being present
        if isinstance(self.command, FunctionType):
            self.command = utils.get_module_and_func_from_function(self.command)

        node_config = {
            'type': 'task',
            'next_node': self.next_node,
            'command': self.command,
            'command_type': self.command_type,
            'command_config': self.command_config,
            'catalog': self.catalog,
            'mode_config': self.mode_config,
            'retry': self.retry,
            'on_failure': self.on_failure
        }
        # The node will temporarily have invalid branch names
        self.node = graph.create_node(name=self.name, step_config=node_config, internal_branch_name='')

    def _fix_internal_name(self):
        pass


class AsIs:
    def __init__(self, name: str, mode_config: Optional[dict] = None, retry: int = 1, on_failure: str = '',
                 next_node: str = '', **kwargs):
        self.name = name
        self.mode_config = mode_config or {}
        self.retry = retry
        self.on_failure = on_failure
        self.next_node = next_node or "success"
        self.additional_kwargs = kwargs or {}
        self.node: Optional[nodes.BaseNode] = None

    def _construct_node(self):
        node_config = {
            'type': 'as-is',
            'next_node': self.next_node,
            'mode_config': self.mode_config,
            'retry': self.retry,
            'on_failure': self.on_failure
        }
        node_config.update(self.additional_kwargs)
        # The node will temporarily have invalid branch names
        self.node = graph.create_node(name=self.name, step_config=node_config, internal_branch_name='')

    def _fix_internal_name(self):
        pass


class Pipeline:
    # A way for the user to define a pipeline
    # TODO: Allow for nodes other than Task, AsIs
    def __init__(
            self, start_at: Union[Task, AsIs],
            name: str = '', description: str = '', max_time: int = defaults.MAX_TIME, internal_branch_name: str = ''):
        self.start_at = start_at
        self.name = name
        self.description = description
        self.max_time = max_time
        self.internal_branch_name = internal_branch_name
        self.dag: Optional[graph.Graph] = None

    def construct(self, steps: List[Task]):
        graph_config: Dict[str, Union[str, int]] = {
            'description': self.description,
            'name': self.name,
            'max_time': self.max_time,
            'internal_branch_name': self.internal_branch_name
        }
        messages: List[str] = []
        for step in steps:
            step._construct_node()
            messages.extend(step.node.validate())  # type: ignore

        if not steps:
            raise Exception('A dag needs at least one step')

        if messages:
            raise Exception(', '.join(messages))

        graph_config['start_at'] = self.start_at.node.name  # type: ignore

        dag = graph.Graph(**graph_config)  # type: ignore
        dag.nodes = [step.node for step in steps]  # type: ignore

        dag.add_terminal_nodes()

        dag.validate()
        self.dag = dag

    def execute(self, configuration_file: str = '', run_id: str = '', tag: str = '', parameters_file: str = '',
                log_level: str = defaults.LOG_LEVEL):

        fileConfig(resource_filename(__name__, 'log_config.ini'))
        logger = logging.getLogger(defaults.NAME)
        logger.setLevel(log_level)

        run_id = utils.generate_run_id(run_id=run_id)
        mode_executor = pipeline.prepare_configurations(
            configuration_file=configuration_file,
            run_id=run_id,
            tag=tag,
            parameters_file=parameters_file)

        mode_executor.execution_plan = defaults.EXECUTION_PLAN.pipeline
        utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

        mode_executor.dag = self.dag
        # Prepare for graph execution
        mode_executor.prepare_for_graph_execution()

        logger.info('Executing the graph')
        mode_executor.execute_graph(dag=mode_executor.dag)

        mode_executor.send_return_code()



import contextlib
import functools
import logging
from typing import Any

import mlflow
from pydantic import BaseModel

from magnus import defaults

logger = logging.getLogger(defaults.NAME)


class BaseExperimentTracker:
    """
    Base Experiment tracker class definition.
    """

    service_name = ''

    class Config(BaseModel):
        pass

    def __init__(self, config: dict = None, **kwargs):  # pylint: disable=unused-argument
        config = config or {}
        self.config = self.Config(**config)

    @property
    def client_context(self):
        return contextlib.nullcontext()

    def log_metric(self, key: str, value: float, step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (float): The value of the metric
            step (int): Optional step at which it was recorded

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        raise NotImplementedError

    def log_parameter(self, key: str, value: Any):
        """
        Logs a parameter in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (any): The value of the metric

        Raises:
            NotImplementedError: Base class, hence not implemented
        """
        pass


class DoNothingTracker(BaseExperimentTracker):
    """
    A Do nothing tracker
    """
    service_name = 'do-nothing'

    def set_metric(self, key: str, value: float, step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (float): The value of the metric
        """
        pass

    def log_parameter(self, key: str, value: Any):
        pass


class MLFlowExperimentTracker(BaseExperimentTracker):
    """
    A MLFlow experiment tracker.

    TODO: Need to set up credentials from secrets
    """
    service_name = 'mlflow'

    class Config(BaseExperimentTracker.Config):
        server_url: str
        autolog: bool = False

    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        self.default_experiment_name = 'Default'
        self.active_run_id = None

        mlflow.set_tracking_uri(self.config.server_url)

        if self.config.autolog:
            mlflow.autolog(log_models=False)

    @functools.cached_property
    def experiment_id(self):
        from magnus.pipeline import global_executor
        experiment_name = self.default_experiment_name

        # If a tag is provided, we should create that as our experiment
        if global_executor.tag:
            experiment_name = global_executor.tag

        experiment = mlflow.get_experiment_by_name(experiment_name)
        if not experiment:
            # Create the experiment and get it.
            experiment = mlflow.create_experiment(experiment_name)
            experiment = mlflow.get_experiment(experiment)

        return experiment.experiment_id

    @functools.cached_property
    def run_name(self):
        from magnus.pipeline import global_executor
        return global_executor.run_id

    @property
    def client_context(self):
        if self.active_run_id:
            return mlflow.start_run(run_id=self.active_run_id,
                                    experiment_id=self.experiment_id, run_name=self.run_name)

        active_run = mlflow.start_run(run_name=self.run_name, experiment_id=self.experiment_id)
        self.active_run_id = active_run.info.run_id
        return active_run

    def set_metric(self, key: str, value: float, step: int = 0):
        """
        Sets the metric in the experiment tracking.

        Args:
            key (str): The key against you want to store the value
            value (Any): The value of the metric
        """
        if not isinstance(value, float):
            msg = (
                f'Only float values are accepted as metrics. Setting the metric {key} as parameter {key}_{step}'
            )
            logger.warning(msg)
            self.log_parameter(key=key+f'_{step}', value=value)
            return

        with self.client_context as _:
            mlflow.log_metric(key, value, step=step or None)

    def get_metric(self, key: str) -> Any:
        """
        Return the metric by the key

        Args:
            key (str): The metric you want to retrieve

        Returns:
            Any: The value of the key
        """
        with self.client_context as client:
            metrics = client.data.metrics
            return metrics.get(key)

    def log_parameter(self, key: str, value: Any):
        with self.client_context as _:
            mlflow.log_param(key, value)

    def get_parameter(self, key: str) -> Any:
        pass
